clear, close all;
Vin = 20;
Vd = 0.7;
Z = 80;
Ts = 25e-6;
L = 430e-6;
C = 22e-6;

% equilibrium point
Vo = 10;
mod.xe = [(Vo + Vin + Vd)*Vo/(Vin*Z);Vo];
mod.ue = (Vo + Vd)/(Vo + Vin + Vd);
mod.x0 = [(2*Vin + Vd)/Z;Vin] - mod.xe;

% model buck-boost converter
% xp = A*x + B*u + C(x)u + f (f not included)
mod.A0 = [1 -Ts/L;Ts/C 1-Ts/(C*Z)];
mod.B0 = [(Vin + Vd)*Ts/L;0];
mod.B{1} = [0;Ts/L];
mod.B{2} = [-Ts/C;0];

% shift model such that the origin is the equilibrium point
mod.A0 = mod.A0 + [mod.ue'*mod.B{1}';mod.ue'*mod.B{2}'];
mod.B0 = mod.B0 + [mod.xe'*mod.B{1};mod.xe'*mod.B{2}];

% set of allowable states
setX.v = [0 0;2.5 0;2.5 50;0 50]' - mod.xe;
setX.A = vert2con(setX.v')';

% set of allowable inputs
setU.v = [0 1] - mod.ue;
[setU.M, setU.um] = vert2con(setU.v');

%% Domain of attraction method lfl
% pole placement
tic;
lfl.lambda = [0.8,0.9];
lfl.K0 = -place(mod.A0,mod.B0,lfl.lambda);

% determine initial polytope (set P)
[P,~] = eig(mod.A0 + mod.B0*lfl.K0);
lfl.setP.A = (kron(eye(2),[1;-1])*P^(-1))';
lfl.setP.v = con2vert(lfl.setP.A')';

% scale initial polytope (set P)
alpha = sdpvar(1);
cons = [];
for i = 1:size(lfl.setP.v,2)
    cons = [cons;setX.A'*lfl.setP.v(:,i)*alpha <= ones(size(setX.A,2),1)];
    cons = [cons;setU.M*lfl.K0*lfl.setP.v(:,i)*alpha <= setU.um];
end
lfl.t = toc;
options = sdpsettings('solver','linprog','verbose',0);
z = optimize(cons,-alpha,options);
lfl.t = lfl.t + z.solvertime;
lfl.setP.v = lfl.setP.v*value(alpha);
lfl.setP.A = vert2con(lfl.setP.v')';

% determine controller corresponding by initial set P
[lfl.K0,epsilon,t0] = LFLDetContr2(mod,lfl.setP,setU);
lfl.t = lfl.t + t0;
% run algorithm
lfl.s0 = 1;
lfl.sast = 0.05;
lfl.maxiter = 200;
lfl.add_vertices = 0;
lfl.zeta = 2;
[lfl.K,lfl.setP,t0] = doa_quad_model_lfl(lfl.setP,lfl.s0,lfl.sast,...
    lfl.zeta,lfl.maxiter,lfl.add_vertices,setX,setU,mod);
lfl.t = lfl.t + t0;

%% Domain of attraction method qlq
% model
qlq.mod.A0 = mod.A0;
qlq.mod.B0 = mod.B0;
for i = 1:size(mod.B0,1)
    qlq.mod.B{i} = zeros(size(mod.B0));
    for j = 1:size(mod.B0,1)
        qlq.mod.B{i}(j,:) = mod.B{j}(i,:);
    end
end
qlq.setP = setX;
[qlq.P,qlq.F,qlq.t] = doa_quad_model_qfq(qlq.mod,qlq.setP,setU);

%% Sequential quadratic programming
sqp.n = size(setX.v,1);
sqp.m = size(setU.v,1);
sqp.Nc = 5;
sqp.N = 1000;
sqp.Q = blkdiag(1,1);
sqp.R = 0.1;
sqp.x0 = -mod.xe;
sqp.setX = setX;
sqp.setU = setU;
sqp.modelfun = @(x,u)BuckBoostModel(x,mod,[],u);
time(1) = 0;
i = 0;
index = 0;
while (1)
    i = i + 1;
%     [sqp.u(:,(i-1)*sqp.Nc+1:i*sqp.Nc),flag,t0] = initseq_sqp(sqp,zeros(sqp.m*sqp.Nc,1));
%     [sqp.u(:,(i-1)*sqp.Nc+1:i*sqp.Nc),flag,t0] = initseq_sqp(sqp,[],1);
    [sqp.u(:,i:i+sqp.Nc-1),flag,t0] = initseq_sqp(sqp,[],1);
    time(1) = time(1) + t0;
    disp(['problem solved: flag = ',num2str(flag),', i = ',num2str(i)]);
    if (i == 1)
%         sqp.x = zeros(sqp.n,ceil(sqp.N/sqp.Nc)*sqp.Nc+1);
        sqp.x = zeros(sqp.n,sqp.N+1);
        sqp.x(:,1) = sqp.x0;
    end
    sqp.x(:,i+1) = sqp.modelfun(sqp.x(:,i),sqp.u(:,i));
%     if (lfl.setP.A'*sqp.x(:,i+1) <= ones(size(lfl.setP.A,2),1)) % lfl
%     method
    if (sqp.x(:,i+1)'*qlq.P*sqp.x(:,i+1) <= 1) % qlq method
        index = i;
        break;
    end
    sqp.x0 = sqp.x(:,i+1);
end
sqp.N = index;
sqp.x = sqp.x(:,1:index+1);
sqp.u = sqp.u(:,1:index);
t = 0:Ts:Ts*sqp.N;
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, Ellipse_plot(qlq.P,zeros(2,1),[],'Orange','fill');
% hold on, plot(Polyhedron(lfl.setP.v'),'alpha',0.35,'color','Orange');
% plot(0,0,'r*');
plot(sqp.x(1,:),sqp.x(2,:),'-o','color','Yellow'), grid;
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}', '\bf{D}', 'origin', 'x');
% xlim([-0.2 2]), ylim([-11 6]);
figure, stairs(t,sqp.x(1,:)), hold on, stairs(t,sqp.x(2,:))
figure, stairs(t(1:end-1),sqp.u,'-o'), grid;
xlabel('time [s]'), ylabel('u');

%% Successive linearization
sln.x0 = -mod.xe;
sln.n = size(mod.A0,1);
sln.m = size(mod.B0,2);
sln.N = 1000;
sln.Nc = 5;
sln.x(:,1) = sln.x0;
sln.Q = blkdiag(1,1);
sln.R = 0;
sln.setX = setX;
sln.setU = setU;
sln.modelfun = @(x,u)BuckBoostModel(x,mod,[],u);
[sln.x,sln.u,sln.N,time(2)] = initseq_sln(sln,mod,qlq.P,1); % qlq method
% [sln.x,sln.u,sln.N,time(2)] = initseq_sln(sln,mod,lfl.setP,1); % lfl
% method
t = 0:Ts:Ts*sln.N;
figure, plot(t,sln.x(1,:)+mod.xe(1)), hold on, plot(t,sln.x(2,:)+mod.xe(2)), grid;
figure, stairs(t(1:end-1),sln.u+mod.ue), grid;
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, Ellipse_plot(qlq.P,zeros(2,1),[],'Orange','fill');
% hold on, plot(Polyhedron(lfl.setP.v'),'alpha',0.35,'color','Orange');
% plot(0,0,'r*')
plot(sln.x(1,:),sln.x(2,:),'-o','color','Yellow'), grid;
% xlim([-0.2 1.8]), ylim([-11 6]);
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}', '\bf{D}', 'origin', 'x');
figure, stairs(t(1:end-1),sln.u,'-o'), grid;
xlabel('time [s]'), ylabel('u');
%% Linear parameter varying MPC
lpv.Nhat = 5;
lpv.p = 1;
lpv.x0 = -mod.xe;
lpv.P = blkdiag(3,1,1);
lpv.Q = blkdiag(3,1,1);
lpv.R = 0.1;
lpv.N = 1000;
[lpv.x,lpv.u,lpv.du,time(3)] = initseq_lpv_mpc_v2(mod,lpv,setX,setU,qlq.P); % qlq method
% [lpv.x,lpv.u,time(3)] = initseq_lpv_mpc_v4(mod,lpv,setX,setU,lfl.setP); %
% lfl method
lpv.N = size(lpv.x,2)-1;
t = 0:Ts:Ts*lpv.N;
figure, plot(t,lpv.x(1,:)+mod.xe(1)), hold on, plot(t,lpv.x(2,:)+mod.xe(2)), grid;
figure, stairs(t(1:end),lpv.u+mod.ue), grid;
figure, stairs(t(1:end),lpv.u,'-o'), grid;
xlabel('time [s]'), ylabel('u');
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, Ellipse_plot(qlq.P,zeros(2,1),[],'Orange','fill');
% hold on, plot(Polyhedron(lfl.setP.v'),'alpha',0.35,'color','Orange');
% plot(0,0,'r*');
plot(lpv.x(1,:),lpv.x(2,:),'-o','color','Yellow'), grid;
% xlim([-0.2 2.3]), ylim([-11 6]);
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}', '\bf{D}', 'origin', 'x');

%%
if (Vo == 10)
    time = [0.2954469 0.6575315 0.4414434;...
            0.2933099 0.2001552 0.5337126];
    time = time + [lfl.t;qlq.t];
    N = [6 29 11;8 14 13];
elseif (Vo == 40)
    time = [0.2356797 0.1001291 0.1304627;...
            0.888606 0.7649099 1.3835638];
    time = time + [lfl.t;qlq.t];
    N = [11 5 4;36 55 51];
else
    error('invalid output reference voltage');
end
X = categorical({'nonlinear MPC (SQP)','successive lin. MPC','LPV-MPC'});
yyaxis left
figure, bar(X,N);
ylabel('length initial sequence')
yyaxis right
hold on, bar(X,time/max(time,[],'all'),0.4);
ylabel('relative time')
grid;
legend('size init. seq.','size init. seq.','relative time','location','best');