clear, close all;
vin = 240;
p = 6;
Rs = 1.5;
lambda_pm = 0.05;
Lq = 1.1e-4;
Ld = 3.5e-4;
JL = 0.019;
DL = 0.12;
TL0 = 30;
Ts = 1e-4;
omega0 = 140;

% equilibrium point of the pmsm
mod.xe(1,1) = 0;
mod.xe(2,1) = 2*(DL*omega0 + TL0)/(sqrt(3)*p*lambda_pm);
mod.xe(3,1) = omega0;
mod.ue(1,1) = -p*Lq*omega0*mod.xe(2)/3;
mod.ue(2,1) = (Rs*mod.xe(2) + sqrt(3)*p*lambda_pm*omega0)/3;

% model pmsm
mod.A0 = [1-Rs*Ts/Ld 0 0;0 1-Rs*Ts/Lq -sqrt(3)*p*lambda_pm*Ts/Lq;...
    0 sqrt(3)*p*lambda_pm*Ts/(2*JL) 1-DL*Ts/JL];
mod.A{1} = zeros(3);
mod.A{1}(3,2) = p*Lq*Ts/Ld;
mod.A{2} = zeros(3);
mod.A{2}(3,1) = -p*Ld*Ts/Lq;
mod.A{3} = zeros(3);
mod.A{3}(1,2) = p*(Ld-Lq)*Ts/(2*JL);
mod.B0 = [3*Ts/Ld 0;0 3*Ts/Lq;0 0];

% shift model such that the origin is the equilibrium point
mod.A0 = mod.A0 + [mod.xe'*mod.A{1};mod.xe'*mod.A{2};mod.xe'*mod.A{3}] + ...
    [mod.xe'*mod.A{1}';mod.xe'*mod.A{2}';mod.xe'*mod.A{3}'];

% set of allowable states
setX.v = interval2vert(kron([-1 1],[210;210;150]))' - mod.xe;
setX.A = vert2con(setX.v')';

% set of allowable inputs
th = linspace(0,2*pi,9);
setU.v = (vin/sqrt(3))*[cos(th(1:end-1));sin(th(1:end-1))] - mod.ue;
[setU.M, setU.um] = vert2con(setU.v');

%% Linear feedback using quadratic Lyapunov function
% determine initial proposed DOA
P = sdpvar(3);
Q = sdpvar(3);
R = sdpvar(2);
cons = [];
for i = 1:size(setX.A,2)
    cons = [cons;[1 setX.A(:,i)';setX.A(:,i) P] >= eps];
end
cons = [cons;[mod.A0'*P*mod.A0-P+Q mod.A0'*P*mod.B0;...
    mod.B0'*P*mod.A0 mod.B0'*P*mod.B0+R] >= 0];
cons = [cons;P >= eps;Q >= 0; R >= eps];
obj = trace(P);
options = sdpsettings('solver','mosek','verbose',0);
z = optimize(cons,obj,options);
lfq.t = z.solvertime;
lfq.setD.P = value(P);

% determine polytope P
N = 12;
th = linspace(0,pi,N+1);
xi = [];
for i = 1:N
    for j = 1:N
        xi = [xi lfq.setD.P^(-0.5)*[cos(th(i));sin(th(i))*cos(2*th(j));sin(th(i))*sin(2*th(j))]];
    end
end
lfq.setP.v = xi;
lfq.setP.A = vert2con(xi')';

gamma = 0;
for i = 1:size(lfq.setP.A,2)
    gamma = max(gamma,lfq.setP.A(:,i)'*lfq.setD.P^(-1)*lfq.setP.A(:,i));
end
lfq.setD.P = lfq.setD.P*value(gamma);
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(lfq.setP.v'),'alpha',0.35,'color','Orange');
Ellipse_plot(lfq.setD.P,zeros(3,1));
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
legend('\bf{X}', '\bf{P}', '\partial\bf{D}_0');
xlim([-220 220]), ylim([-35 35]), zlim([-11 11]);

% apply algorithm in order to find feedback gain and DOA
lfq.delta = 0.01;
lfq.imax = 20;
[lfq.K,lfq.setP,t0] = doa_quad_model_lfq(mod,lfq.setD.P,...
    lfq.setP,setU,lfq.delta,lfq.imax);
lfq.t = lfq.t + t0;
gamma = 0;
for i = 1:size(lfq.setP.A,2)
    gamma = max(gamma,lfq.setP.A(:,i)'*lfq.setD.P^(-1)*lfq.setP.A(:,i));
end
lfq.setD.P = lfq.setD.P*gamma;

%% Sequential quadratic programming

sqp.n = size(setX.v,1);
sqp.m = size(setU.v,1);
sqp.Nc = 5;
sqp.N = 2e4;
sqp.Q = blkdiag(1,1,10);
sqp.R = blkdiag(0.1,0.1);
sqp.x0 = -mod.xe;
sqp.setX = setX;
sqp.setU = setU;
sqp.modelfun = @(x,u)PMSMModel(x,mod,[],u);
sqp.x = zeros(sqp.n,ceil(sqp.N/sqp.Nc)*sqp.Nc);
sqp.x(:,1) = -mod.xe;
sqp.u = zeros(sqp.m,ceil(sqp.N/sqp.Nc)*sqp.Nc);
fiter = 1;
j = 1;
time(1) = 0;
rho = inf;
for k = 1:sqp.N
    if (k > 1)
        sqp.x0 = sqp.x(:,k);
    end
    while(1)
        if (k > 1 && fiter == 1)
            [u,flag(j),t0] = initseq_sqp(sqp,reshape(u,sqp.m*sqp.Nc,1),1);
        elseif (fiter == 1)
            [u,flag(j),t0] = initseq_sqp(sqp,[],1);
        else
            u0 = 100*rand(sqp.m*sqp.Nc,1);
            [u,flag(j),t0] = initseq_sqp(sqp,u0,1);
        end
        time(1) = time(1) + t0;
        disp(['problem solved: flag = ',num2str(flag(j)),', k = ',num2str(k),' (',num2str(rho),')']);
        if (flag(j) == 0 || flag(j) == 1 || flag(j) == -3)
            sqp.u(:,k) = u(:,1);
            if (k == 1)
                sqp.x = zeros(sqp.n,sqp.N+1);
                sqp.x(:,1) = sqp.x0;
            end
            sqp.x(:,k+1) = sqp.modelfun(sqp.x(:,k),sqp.u(:,k));
            sqp.x0 = sqp.x(:,k+1);
            fiter = 1;
            j = j + 1;
            break;
        end
        j = j + 1;
    end
    rho = sqp.x(:,k+1)'*lfq.setD.P*sqp.x(:,k+1);
    if (rho <= 1)
        break;
    end
end
sqp.N = k;
sqp.x = sqp.x(:,1:sqp.N+1);
sqp.u = sqp.u(:,1:sqp.N);
t = 0:Ts:Ts*sqp.N;
figure, Ellipse_plot(lfq.setD.P,zeros(3,1)',[],'Yellow','fill')
hold on, plot3(sqp.x(1,:),sqp.x(2,:),sqp.x(3,:)), hold on, plot3(0,0,0,'r*');
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
legend('\bf{D}', 'x', 'origin');
figure, stairs(t(1:end-1),sqp.u(1,:)), hold on, stairs(t(1:end-1),sqp.u(2,:)), grid;
xlabel('time [s]');
legend('u^{(1)}', 'u^{(2)}');

%% Successive linearization
sln.x0 = -mod.xe;
sln.n = size(mod.A0,1);
sln.m = size(mod.B0,2);
sln.N = 2e4;
sln.Nc = 5;
sln.x(:,1) = sln.x0;
sln.Q = blkdiag(1,1,10);
sln.R = zeros(2);
sln.setX = setX;
sln.setU = setU;
sln.modelfun = @(x,u)PMSMModel(x,mod,[],u);
[sln.x,sln.u,sln.N,time(2)] = initseq_sln(sln,mod,lfq.setD.P,2);
t = 0:Ts:Ts*sln.N;
figure, stairs(t,sln.x(1,:)+mod.xe(1)), hold on, stairs(t,sln.x(2,:)+mod.xe(2)), grid;
stairs(t,sln.x(3,:)+mod.xe(3))
figure, stairs(t(1:end-1),sln.u(1,:)+mod.ue(1)), hold on, stairs(t(1:end-1),sln.u(2,:)+mod.ue(2)), grid;
figure, stairs(t(1:end-1),sln.u'), grid;
legend('u^{(1)}', 'u^{(2)}');
xlabel('time [s]')
% figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
figure, Ellipse_plot(lfq.setD.P,zeros(3,1),[],'Yellow','fill');
hold on, plot3(sln.x(1,:),sln.x(2,:),sln.x(3,:),'-','color','Green'), grid on;
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
legend('\bf{D}', 'x');

%% Linear parameter varying MPC
lpv.Nhat = 4;
lpv.p = 1;
lpv.x0 = -mod.xe;
lpv.P = blkdiag(1,1,10,1,1);
lpv.Q = blkdiag(1,1,10,1,1);
lpv.R = blkdiag(0.1,0.1);
lpv.N = 2e4;
[lpv.x,lpv.u,lpv.du,time(3)] = initseq_lpv_mpc_v2(mod,lpv,setX,setU,lfq.setD.P);
lpv.N = size(lpv.x,2)-1;
t = 0:Ts:Ts*lpv.N;
figure, plot(t,lpv.x(1,:)+mod.xe(1)), hold on, plot(t,lpv.x(2,:)+mod.xe(2)), grid;
plot(t,lpv.x(3,:)+mod.xe(3));
figure, stairs(t,lpv.u(1,:)+mod.ue(1)), hold on, plot(t,lpv.u(2,:)+mod.xe(2)), grid;
figure, stairs(t,lpv.u'), grid;
legend('u^{(1)}', 'u^{(2)}');
xlabel('time [s]')
% figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
figure, Ellipse_plot(lfq.setD.P,zeros(3,1),[],'Yellow','fill');
hold on, plot3(lpv.x(1,:),lpv.x(2,:),lpv.x(3,:),'-','color','Green'), plot3(0,0,0,'r*'), grid on;
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
legend('\bf{D}', 'x', 'origin');

%%
time = time + lfq.t;
N = [sqp.N sln.N lpv.N];
X = categorical({'nonlinear MPC (SQP)','successive lin. MPC','LPV-MPC'});
figure;
yyaxis left
bar(X,N);
ylabel('length initial sequence');
yyaxis right
hold on, bar(X,time/max(time,[],'all'),0.4);
ylabel('relative time')
grid;
legend('size init. seq.','relative time','location','best');