function [x,u,du,time] = initseq_lpv_mpc_v2(mod,lpv,setX,setU,varargin)
% This function applies the LPV-MPC method to the LPV system defined by
% mod.
% [x,u,du,time] = initseq_lpv_mpc_v2(mod,lpv,setX,setU)
% [x,u,du,time] = initseq_lpv_mpc_v2(mod,lpv,setX,setU,setP)
% [x,u,du,time] = initseq_lpv_mpc_v2(mod,lpv,setX,setU,P)
% inputs:
% mod : structure consisting of the matrices that form the quadratic model
% of the system (A0, A{}, B0, B{})
% lpv : structure consisting of all the parameters needed to apply this
% method (Nhat (control horizon), p (norm of obj. function), x0 (initial
% state), P (weighting matrix on final state), Q (weighting matrix on
% states, R (weighting matrix on inputs), N (size of the initial sequence))
% setX : set of allowable states (A*x <= 1)
% setU : set of allowable inputs (M*u <= um)
% setP : domain of attraction of local controller (A*x <= 1)
% P : domain of attraction of local controller (x'*P*x <= 1)
% 
% outputs:
% x : states of the system corresponding to the initial control sequence
% u : initial control inputs
% du : change in the initial control inputs
% time : computational time of this method
% 
% Rik Koch (TU Delft, 2021)

rmpath('C:\Program Files\Mosek\9.2\toolbox\R2015a');
q = size(setX.v,2);
n = size(setX.v,1);
m = size(setU.M,2);
mu = m+n;
N = lpv.Nhat;

% construction of the system matrices
Psi = cell(q,1);
for i = 1:q
    Psi{i} = [mod.A0 mod.B0;zeros(m,n) eye(m)];
    if (isfield(mod,'A'))
        for j = 1:n
            Psi{i}(j,1:n) = Psi{i}(j,1:n) + setX.v(:,i)'*mod.A{j};
        end
    end
    if (isfield(mod,'B'))
        for j = 1:n
            Psi{i}(j,n+1:n+m) = Psi{i}(j,n+1:n+m) + setX.v(:,i)'*mod.B{j};
        end
    end
end
Gamma = [zeros(n,m);eye(m)];
A = [setX.A' zeros(size(setX.A,2),m);zeros(size(setU.M,1),n) setU.M];
b = [ones(size(setX.A,2),1);setU.um];

if (nargin > 4)
    if (isstruct(varargin{1}))
        setP = varargin{1};
    else
        P = varargin{1};
    end
end
qi = zeros(N,1);
qi(1) = 1;
for i = 1:N-1
    qi(i+1) = qi(i) + q^i;
end
% construction of the linear programming problem
As = zeros(mu*qi(end),mu);
Bs = zeros(mu*qi(end),m*(N-1)*q);
As(1:mu,:) = eye(mu);
for j1 = 1:q
    ind1 = (qi(1)+j1-1)*mu+1:(qi(1)+j1)*mu;
    ind2 = (j1-1)*m+1:j1*m;
    As(ind1,:) = Psi{j1};
    Bs(ind1,ind2) = Gamma;
    for j2 = 1:q
        ind1 = (qi(2)+(j1-1)*q+j2-1)*mu+1:(qi(2)+(j1-1)*q+j2)*mu;
        ind2 = (j1-1)*m+1:j1*m;
        ind3 = (q+j2-1)*m+1:(q+j2)*m;
        As(ind1,:) = Psi{j2}*Psi{j1};
        Bs(ind1,ind2) = Psi{j2}*Gamma;
        Bs(ind1,ind3) = Gamma;
        for j3 = 1:q
            ind1 = (qi(3)+(j1-1)*q^2+(j2-1)*q+j3-1)*mu+1:(qi(3)+(j1-1)*q^2+(j2-1)*q+j3)*mu;
            ind2 = (j1-1)*m+1:j1*m;
            ind3 = (q+j2-1)*m+1:(q+j2)*m;
            ind4 = (2*q+j3-1)*m+1:(2*q+j3)*m;
            As(ind1,:) = Psi{j3}*Psi{j2}*Psi{j1};
            Bs(ind1,ind2) = Psi{j3}*Psi{j2}*Gamma;
            Bs(ind1,ind3) = Psi{j3}*Gamma;
            Bs(ind1,ind4) = Gamma;
            for j4 = 1:q
                if (N < 5)
                    break;
                end
                ind1 = (qi(4)+(j1-1)*q^3+(j2-1)*q^2+(j3-1)*q+j4-1)*mu+1:(qi(4)+(j1-1)*q^3+(j2-1)*q^2+(j3-1)*q+j4)*mu;
                ind2 = (j1-1)*m+1:j1*m;
                ind3 = (q+j2-1)*m+1:(q+j2)*m;
                ind4 = (2*q+j3-1)*m+1:(2*q+j3)*m;
                ind5 = (3*q+j4-1)*m+1:(3*q+j4)*m;
                As(ind1,:) = Psi{j4}*Psi{j3}*Psi{j2}*Psi{j1};
                Bs(ind1,ind2) = Psi{j4}*Psi{j3}*Psi{j2}*Gamma;
                Bs(ind1,ind3) = Psi{j4}*Psi{j3}*Gamma;
                Bs(ind1,ind4) = Psi{j4}*Gamma;
                Bs(ind1,ind5) = Gamma;
            end
        end
    end
end

G1 = zeros(mu*(qi(end)+1),(mu+m)*N+mu);
G2 = zeros(m*(1+(N-1)*q),(mu+m)*N+mu);
G1(1:2*mu,1:2*mu) = eye(2*mu);
G2(1:m,mu*(N+1)+1:mu*(N+1)+m) = eye(m);
for i = 1:N-1
    ind1 = (1+qi(i))*mu+1:(1+qi(i+1))*mu;
    ind2 = ((i+1)*mu+1:(i+2)*mu);
    G1(ind1,ind2) = kron(ones(q^i,1),eye(mu));
    ind1 = (1+(i-1)*q)*m+1:(1+i*q)*m;
    ind2 = (mu*(N+1)+i*m+1:mu*(N+1)+(i+1)*m);
    G2(ind1,ind2) = kron(ones(q,1),eye(m));
end
Qhat = zeros(qi(end)*mu,qi(end)*mu,'single');
Qhat(1:end,1:end) = blkdiag(kron(eye(qi(end-1)),lpv.Q),kron(eye(q^(N-1)),lpv.P));
Rhat = zeros((N-1)*q*m,(N-1)*q*m);
Rhat(1:end,1:end) = kron(eye((N-1)*q),lpv.R);
Ahat = zeros(qi(end)*size(A,1),qi(end)*size(A,2),'single');
Ahat(1:end,1:end) = kron(eye(qi(end)),A);
bhat = zeros(qi(end)*size(b,1),size(b,2));
bhat(1:end,1:end) = kron(ones(qi(end),1),b);

x = lpv.x0;
k = 1;
time = 0;
while(1)
    % construct Psi0
    Psi0 = [mod.A0 mod.B0;zeros(m,n) eye(m)];
    if (isfield(mod,'A'))
        for j = 1:n
            Psi0(j,1:n) = Psi0(j,1:n) + x(:,k)'*mod.A{j};
        end
    end
    if (isfield(mod,'B'))
        for j = 1:n
            Psi0(j,n+1:n+m) = Psi0(j,n+1:n+m) + x(:,k)'*mod.B{j};
        end
    end
    % determine initial state
    if (k > 1)
        y0 = [x(:,k);u(:,k)];
    elseif (nargin > 5)
        u(:,k) = varargin{2};
        y0 = [x(:,k);u(:,k)];
    else
        switch lpv.p
            case 1
                Aieq = [lpv.P(1:n,1:n)*Psi0(1:n,n+1:n+m) -eye(n);...
                    -lpv.P(1:n,1:n)*Psi0(1:n,n+1:n+m) -eye(n);...
                    setX.A'*Psi0(1:n,n+1:n+m) zeros(size(setX.A,2),n);...
                    setU.M zeros(size(setU.M,1),n)];
                bieq = [-lpv.P(1:n,1:n)*Psi0(1:n,1:n)*lpv.x0;...
                    lpv.P(1:n,1:n)*Psi0(1:n,1:n)*lpv.x0;...
                    ones(size(setX.A,2),1)-setX.A'*Psi0(1:n,1:n)*lpv.x0;setU.um];
                f = [zeros(m,1);ones(n,1)];
            case inf
                Aieq = [lpv.P(1:n,1:n)*Psi0(1:n,n+1:n+m) -ones(n,1);...
                    -lpv.P(1:n,1:n)*Psi0(1:n,n+1:n+m) -ones(n,1);...
                    setX.A'*Psi0(1:n,n+1:n+m) zeros(size(setX.A,2),1);...
                    setU.M zeros(size(setU.M,1),1)];
                bieq = [-lpv.P(1:n,1:n)*Psi0(1:n,1:n)*lpv.x0;...
                    lpv.P(1:n,1:n)*Psi0(1:n,1:n)*lpv.x0;...
                    ones(size(setX.A,2),1)-setX.A'*Psi0(1:n,1:n)*lpv.x0;setU.um];
                f = [zeros(m,1);1];
            otherwise
                error('Invalid norm. Change the value of lpv.p into 1 or inf.');
        end
        options = optimoptions('linprog','Algorithm','dual-simplex','Display','off');
        [u0,~,flag] = linprog(f,Aieq,bieq,[],[],[],[],options);
        u(:,k) = u0(1:m);
        y0 = [x(:,k);u(:,k)];
        if (flag ~= 1)
            if (flag == 3)
                warning('Solution not feasible wrt. absolute tolerance');
            else
                error('No feasible solution found');
            end
        end
    end
    
    Aieq = [[zeros(mu,m);Qhat*As*Gamma] [zeros(mu,(N-1)*q*m);Qhat*Bs] -G1;...
            -[zeros(mu,m);Qhat*As*Gamma] -[zeros(mu,(N-1)*q*m);Qhat*Bs] -G1;...
            [lpv.R;zeros((N-1)*q*m,m)] [zeros(m,(N-1)*q*m);Rhat] -G2;...
            -[lpv.R;zeros((N-1)*q*m,m)] -[zeros(m,(N-1)*q*m);Rhat] -G2;...
            Ahat*As*Gamma Ahat*Bs zeros(qi(end)*size(A,1),N*(mu+m)+mu)];
    bieq = [-lpv.Q*y0;-Qhat*As*Psi0*y0;lpv.Q*y0;Qhat*As*Psi0*y0;zeros(2*m*((N-1)*q+1),1);bhat-Ahat*As*Psi0*y0];
    f = [zeros(m,1);zeros((N-1)*q*m,1);ones(N*(mu+m)+mu,1)];
    options = optimoptions('linprog','Algorithm','dual-simplex','Display','off');
    tic;
    [s,~,flag] = linprog(f,double(Aieq),double(bieq),[],[],[],[],options);
    time = time + toc;
    if (isempty(s))
        addpath('C:\Program Files\Mosek\9.2\toolbox\R2015a');
        error('Linear control problem not solvable');
    end
    disp(['flag: ',num2str(flag),' [k = ',num2str(k),', obj = ',num2str(f'*s),']']);
    k = k + 1;
    du(:,k-1) = s(1:m);
    y1 = Psi0*y0 + Gamma*du(:,k-1);
    x(:,k) = y1(1:n);
    u(:,k) = y1(n+1:n+m);
    if (exist('setP','var'))
        if (max(setP.A'*x(:,k)) <= 1)
            break;
        end
    elseif (exist('P','var'))
        if (x(:,k)'*P*x(:,k) <= 1)
            break;
        end
    end
    if (k >= lpv.N)
        break;
    end
end
addpath('C:\Program Files\Mosek\9.2\toolbox\R2015a');
end