function [x,u,N,time] = initseq_sln(sln,mod,varargin)
% Function that determines the initial control sequence using successive
% linearization for quadratic systems.
% [x,u,N,time] = initseq_sln(sln,mod)
% [x,u,N,time] = initseq_sln(sln,mod,setP)
% [x,u,N,time] = initseq_sln(sln,mod,P)
% inputs:
% mod : structure consisting of the matrices that form the quadratic model
% of the system (A0, A{}, B0, B{})
% sln : structure consisting of all the parameters needed to apply this
% method (Nc (control horizon), x0 (initial state), P (weighting matrix on
% final state), Q (weighting matrix on states, R (weighting matrix on
% inputs), N (size of the initial sequence), n (number of states), m
% (number of inputs), setX (set of allowable states (A*x <= 1)), setU (set
% of allowable inputs (M*u <= um)), modelfun (function of the quadratic
% model)
% setP : domain of attraction of local controller (A*x <= 1)
% P : domain of attraction of local controller (x'*P*x <= 1)
% 
% outputs:
% x : states of the system corresponding to the initial control sequence
% u : initial control inputs
% N : size of the initial control sequence
% time : computational time of this method
% 
% Rik Koch (TU Delft, 2021)

rmpath('C:\Program Files\Mosek\9.2\toolbox\R2015a');
if (nargin > 2 && ~isempty(varargin{1}))
    if (isstruct(varargin{1}))
        setP = varargin{1};
    else
        P = varargin{1};
    end
end
if (nargin > 3)
    mode = varargin{2};
else
    mode = 2;
end
N = sln.N;
Nc = sln.Nc;
n = sln.n;
m = sln.m;
p = size(sln.setX.A,2);
q = size(sln.setU.M,1);
x = zeros(n,N+Nc);
u = zeros(m,N+Nc-1);
x(:,1) = sln.x0;
if (mode == 2)
    [U,~,t0] = initseq_sqp(sln,zeros(m*Nc,1),1);
    time = t0;
    u(:,1:Nc) = reshape(U,m,Nc);
    for i = 1:Nc
        x(:,i+1) = sln.modelfun(x(:,i),u(:,i));
    end
else
    time = 0;
end
A = cell(Nc,1);
B = cell(Nc,1);
for k = mode:N
    F = kron(ones(Nc,1),eye(n));
    G = zeros(n*Nc,m*Nc);
    for i = 1:Nc
        if (mode == 1 && i > 1)
            A{i} = A{i-1};
            B{i} = B{i-1};
        else
            A{i} = mod.A0;
            if (isfield(mod,'A'))
                for j = 1:n
                    A{i}(j,:) = A{i}(j,:) + x(:,k+i-1)'*mod.A{j};
                end
            end
            B{i} = mod.B0;
            if (isfield(mod,'B'))
                for j = 1:n
                    B{i}(j,:) = B{i}(j,:) + x(:,k+i-1)'*mod.B{j};
                end
            end
        end
        for j = 1:i
            F((i-1)*n+1:i*n,:) = F((i-1)*n+1:i*n,:)*A{j};
            Gt = eye(n);
            for l = 1:i-j
                Gt = Gt*A{i-l+1};
            end
            G((i-1)*n+1:i*n,(j-1)*m+1:j*m) = Gt*B{j};
        end
    end
    Aieq = zeros(p*Nc + q*Nc + 3*n*Nc + 1,m*Nc + n*Nc + 1);
    bieq = zeros(p*Nc + q*Nc + 3*n*Nc + 1,1);
    Aieq(1:p*Nc,1:m*Nc) = kron(eye(Nc),sln.setX.A')*G;
    bieq(1:p*Nc) = ones(p*Nc,1) - kron(eye(Nc),sln.setX.A')*F*x(:,k);
    Aieq(p*Nc+1:(p+q)*Nc,1:m*Nc) = kron(eye(Nc),sln.setU.M);
    bieq(p*Nc+1:(p+q)*Nc) = kron(ones(Nc,1),sln.setU.um);
    Aieq((p+q)*Nc+1:(p+q+n)*Nc,:) = [kron(blkdiag(100,eye(Nc-1)),sln.Q)*G -eye(n*Nc) zeros(n*Nc,1)];
    bieq((p+q)*Nc+1:(p+q+n)*Nc) = -kron(blkdiag(100,eye(Nc-1)),sln.Q)*F*x(:,k);
    Aieq((p+q+n)*Nc+1:(p+q+2*n)*Nc,:) = [-kron(blkdiag(100,eye(Nc-1)),sln.Q)*G -eye(n*Nc) zeros(n*Nc,1)];
    bieq((p+q+n)*Nc+1:(p+q+2*n)*Nc) = kron(blkdiag(100,eye(Nc-1)),sln.Q)*F*x(:,k);
    Aieq((p+q+2*n)*Nc+1:end-1,m*Nc+1:end-1) = -eye(n*Nc);
    Aieq(end,m*Nc+1:end) = [ones(1,n*Nc) -1];
    f = [zeros((m+n)*Nc,1);1];
    options = optimoptions('linprog','Algorithm','dual-simplex','Display','off');
    tic;
    [U,~,flag] = linprog(f,Aieq,bieq,[],[],[],[],options);
    time = time + toc;
    disp(['flag: ',num2str(flag),' [k = ',num2str(k),', obj = ',num2str(U(end)),']']);
    u(:,k:k+Nc-1) = reshape(U(1:m*Nc),[m,Nc]);
    x(:,k+1) = sln.modelfun(x(:,k),u(:,k));
    if (exist('setP','var'))
        if (max(setP.A'*x(:,k+1)) <= 1)
            break;
        end
    elseif (exist('P','var'))
        if (x(:,k+1)'*P*x(:,k+1) <= 1)
            break;
        end
    end
    for i = 1:Nc-1
        x(:,k+i+1) = sln.modelfun(x(:,k+i),u(:,k+i));
    end
end
N = k;
x = x(:,1:N+1);
u = u(:,1:N);
addpath('C:\Program Files\Mosek\9.2\toolbox\R2015a');
end