function [u,flag,time,varargout] = initseq_sqp(sqp,varargin)
% Function that determines consecutive control inputs using sequential
% quadratic programming.
% [u,flag,time] = initseq_sqp(sqp)
% [u,flag,time] = initseq_sqp(sqp,u0)
% [u,flag,time] = initseq_sqp(sqp,[],yalmipsol)
% [u,flag,time] = initseq_sqp(sqp,u0,yalmipsol)
% inputs:
% mod : structure consisting of the matrices that form the quadratic model
% of the system (A0, A{}, B0, B{})
% sln : structure consisting of all the parameters needed to apply this
% method (Nc (control horizon), x0 (initial state), P (weighting matrix on
% final state), Q (weighting matrix on states, R (weighting matrix on
% inputs), N (size of the initial sequence), n (number of states), m
% (number of inputs), setX (set of allowable states (A*x <= 1)), setU (set
% of allowable inputs (M*u <= um)), modelfun (function of the quadratic
% model)
% u0 : initial guess of the control input (warm sequence)
% yalmipsol : if 1, solve the programming problem using the yalmip toolbox
% 
% outputs:
% x (4th output) : next consecutive state of the system determined using
% sqp
% obj (5th output) : value of the objective function after applying sqp
% u : initial control inputs
% time : computational time of this method
% flag : indicates a valid solution (check documentation of yalmip or
% fmincon)
% 
% Rik Koch (TU Delft, 2021)

if (nargin > 1 && ~isempty(varargin{1}))
    u0 = varargin{1};
else
    u0 = zeros(sqp.m*sqp.Nc,1);
end
if (nargin > 2 && ~isempty(varargin{2}))
    yalmipsol = varargin{2};
else
    yalmipsol = 0;
end
if (yalmipsol == 0)
    options = optimoptions('fmincon','Algorithm','sqp','Display','off');
    tic;
    [u,~,flag,obj] = fmincon(@(u)objsqp(u,sqp),u0,[],[],[],[],[],[],@(u)nonlconsqp(u,sqp),options);
    time = toc;
else
    u = sdpvar(sqp.m*sqp.Nc,1);
    [obj,x] = objsqp(u,sqp,1);
    [c,ceq] = nonlconsqp(u,sqp,1,x{1});
    cons = [c <= zeros(size(c,1),1);ceq <= zeros(size(ceq,1),1)];
    options = sdpsettings('solver','fmincon','verbose',0);
    options.fmincon.Algorithm = 'sqp';
    z = optimize(cons,obj,options);
    if (z.info(1) == 'S')
        u = value(u);
        flag = z.problem;
        time = z.solvertime;
    else
        error('problem not solved');
    end
end
u = reshape(u,sqp.m,sqp.Nc);
if(nargout > 3)
    varargout{1} = value(x{1});
    varargout{2} = value(obj);
end
end