function [K,epsilon,t] = LFLDetContr2(mod,setP,setU)
% This function determines the controller for the quadratic model based on
% a linear Lyapunov function
% corresponding to a specified contractive set.
% input:
% mod: the state matrices of the quadratic model (x_{k+1} = (A0 + A)*x_k +
% (B0 + B)*u_k. A and B must be represented as cell if they exist.
% setP: a matrix representing the contractive set (A'*x <= 1)
% (vertices of the cones)
% M, um: input constraints (M*u <= um)
% xs,us: the equilibrium state and input, respectively
% output:
% K: the controller gain (g(x) = K*(x - xs) + us)
% eps: variable indicating whether the controller is asymptotically stable
% (eps < 1) or not (eps >= 0)
% 
% Rik Koch (TU Delft, 2021)

tic;
n = size(mod.A0,1);
m = size(mod.B0,2);
p = size(setP.A,2);
q = size(setP.v,2);


Eps = LFLDetEps(setP);

% textprogressbar('constructing the constraints ');
Aieq = zeros(n*(n+1)*p^2 + size(setU.um,1)*q + 1,m*n+1);
bieq = zeros(n*(n+1)*p^2 + size(setU.um,1)*q + 1,1);
iter = 0;
for i = 1:p
    x = Eps{i};
    for j = 1:p
        for k1 = 2:size(x,2) % loop over v1
            for k2 = 1:size(x,2) % loop over v2
                A = mod.A0;
                B = mod.B0;
                for k = 1:n
                    if(isfield(mod,'A'))
                        A(k,:) = A(k,:) + x(:,k2)'*mod.A{k};
                    end
                    if(isfield(mod,'B'))
                        B(k,:) = B(k,:) + x(:,k2)'*mod.B{k};
                    end                    
                end
                Aieq(iter+1,:) = [kron(x(:,k1)',setP.A(:,j)'*B) -setP.A(:,i)'*x(:,k1)];
                bieq(iter+1) = -setP.A(:,j)'*A*x(:,k1);
                iter = iter+1;
            end
        end
%         textprogressbar(100*((i-1)*p+j)/p^2);
%         pause(1e-6);
    end
end
for i = 1:q
    Aieq(iter+(i-1)*size(setU.um,1)+1:iter+i*size(setU.um,1),:) = [kron(setP.v(:,i)',setU.M) zeros(size(setU.um,1),1)];
    bieq(iter+(i-1)*size(setU.um,1)+1:iter+i*size(setU.um,1)) = setU.um;
end
% textprogressbar('done');
Aieq(end:end) = -1;
bieq(end) = -eps;
f = [zeros(m*n,1);1];
options = optimoptions('linprog','Algorithm','dual-simplex','Display','off');
[x,~,flag] = linprog(f,Aieq,bieq,[],[],[],[],options);
t = toc;
if (flag == 1)
    epsilon = x(end);
    K = reshape(x(1:end-1),[m,n]);
else
    epsilon = inf;
    K = [];
end
end