clear, close all;
Vin = 20;
Vd = 0.7;
Z = 80;
Ts = 25e-6;
L = 430e-6;
C = 22e-6;

% equilibrium point
Vo = 10;
mod.xe = [(Vo + Vin + Vd)*Vo/(Vin*Z);Vo];
mod.ue = (Vo + Vd)/(Vo + Vin + Vd);

% model buck-boost converter
% xp = A*x + B*u + C(x)u + f (f not included)
mod.A0 = [1 -Ts/L;Ts/C 1-Ts/(C*Z)];
mod.B0 = [(Vin + Vd)*Ts/L;0];
mod.B{1} = [0;Ts/L];
mod.B{2} = [-Ts/C;0];

% shift model such that the origin is the equilibrium point
mod.A0 = mod.A0 + [mod.ue'*mod.B{1}';mod.ue'*mod.B{2}'];
mod.B0 = mod.B0 + [mod.xe'*mod.B{1};mod.xe'*mod.B{2}];

% set of allowable states
setX.v = [0 0;2.5 0;2.5 50;0 50]' - mod.xe;
setX.A = vert2con(setX.v')';

% set of allowable inputs
setU.v = [0 1] - mod.ue;
[setU.M, setU.um] = vert2con(setU.v');
%% random sampling approach
% linear quadratic controller
tic;
rsa.Q = blkdiag(1/mod.xe(1)^2,1/mod.xe(2)^2);
rsa.R = 1/mod.ue^2;
[rsa.P0,~,rsa.K] = dare(mod.A0,mod.B0,rsa.Q,rsa.R);
rsa.K = -rsa.K;

% determine an estimation of the DOA;
modelfun = @(x)BuckBoostModel(x,mod,rsa);
rsa.delta = 0.001;
rsa.past = 0.999;
rsa.mu = 0.001;
rsa.tau = 0.001;
t(1) = toc;
[rsa.P0,t0] = random_sampling_doa(setX,setU,rsa.P0,modelfun,rsa.delta,...
    rsa.past,rsa.mu,rsa.tau);
t(1) = t(1) + t0;
vol(1) = pi*prod(1./sqrt(eig(rsa.P0)))/volume(Polyhedron(setX.v'));

%% Linear feedback using linear Lyapunov function
% check whether setX is sufficient for finding a stabilising controller
[K,epsilon,t0] = LFLDetContr2(mod,setX,setU);
t(2) = t0;
% pole placement
tic;
% lfl.lambda = [0.8,0.9];
lfl.lambda = [0.7 + 0.5i,0.7 - 0.5i];
lfl.K0 = -place(mod.A0,mod.B0,lfl.lambda);

% determine initial polytope (set P)
[P,~] = eig(mod.A0 + mod.B0*lfl.K0);
if (sum(imag(lfl.lambda) == 0) == 0)
    rho = real(lfl.lambda(1));
    beta = abs(imag(lfl.lambda(1)));
    N = ceil(2*pi/beta);
    W = zeros(N,2);
    for k = 1:N
        W(k,:) = [cos((2*k-1)*pi/N) sin((2*k-1)*pi/N)];
    end
    w = cos(pi/N)*ones(N,1);
    P = [real(P(:,1)) imag(P(:,1))];
    lfl.setP.A = (W*P^(-1)./w)';
    lfl.setP.v = con2vert(lfl.setP.A')';
else
    lfl.setP.A = (kron(eye(2),[1;-1])*P^(-1))';
    lfl.setP.v = con2vert(lfl.setP.A')';
end
figure, plot(polyshape(setX.v')), grid;
hold on, plot(Polyhedron(lfl.setP.v'),'alpha',0.35,'color','Orange');
plot(0,0,'r*');
xlim([-0.4 0.4]), ylim([-3 3]);
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}','\bf{P}_0');

% scale initial polytope (set P)
alpha = sdpvar(1);
cons = [];
for i = 1:size(lfl.setP.v,2)
    cons = [cons;setX.A'*lfl.setP.v(:,i)*alpha <= ones(size(setX.A,2),1)];
    cons = [cons;setU.M*lfl.K0*lfl.setP.v(:,i)*alpha <= setU.um];
end
t(2) = t(2) + toc;
options = sdpsettings('solver','linprog','verbose',0);
z = optimize(cons,-alpha,options);
t(2) = t(2) + z.solvertime;
lfl.setP.v = lfl.setP.v*value(alpha);
lfl.setP.A = vert2con(lfl.setP.v')';

figure, plot(polyshape(setX.v')), grid;
hold on, plot(Polyhedron(lfl.setP.v'),'alpha',0.35,'color','Orange');
plot(0,0,'r*');
xlim([-0.2 0.2]), ylim([-1.5 1.5]);
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}','\bf{P}_0');

% determine controller corresponding by initial set P
[lfl.K0,epsilon,t0] = LFLDetContr2(mod,lfl.setP,setU);
t(2) = t(2) + t0;
% run algorithm
lfl.s0 = 1;
lfl.sast = 0.05;
lfl.maxiter = 200;
lfl.add_vertices = 0;
lfl.zeta = 2;
% lfl.zeta = (sqrt(5)+1)/2;
% lfl.zeta = fibonacci(3:15);
[lfl.K,lfl.setP,t0] = doa_quad_model_lfl(lfl.setP,lfl.s0,lfl.sast,...
    lfl.zeta,lfl.maxiter,lfl.add_vertices,setX,setU,mod);
t(2) = t(2) + t0;
vol(2) = volume(Polyhedron(lfl.setP.v'))/volume(Polyhedron(setX.v'));

%% Linear feedback using quadratic Lyapunov function
% determine initial proposed DOA
tic;
P = sdpvar(2);
Q = sdpvar(2);
R = sdpvar(1);
cons = [];
if (Vo == 10)
    lfq.setX.v = [0 5;0 15;2.5 15;2.5 5]'-mod.xe;
else
    lfq.setX.v = setX.v;
end
lfq.setX.A = vert2con(lfq.setX.v')';
for i = 1:size(setX.A,2)
    cons = [cons;[1 lfq.setX.A(:,i)';lfq.setX.A(:,i) P] >= eps];
end
cons = [cons;[mod.A0'*P*mod.A0-P+Q mod.A0'*P*mod.B0;...
    mod.B0'*P*mod.A0 mod.B0'*P*mod.B0+R] >= 0];
cons = [cons;P >= eps;Q >= 0; R >= eps];
obj = trace(P);
t(3) = toc;
options = sdpsettings('solver','mosek','verbose',0);
z = optimize(cons,obj,options);
t(3) = t(3) + z.solvertime;
lfq.setD.P = value(P);

% determine polytope P
tic;
N = 12;
th = linspace(0,2*pi,N+1);
xi = [];
for i = 1:N
    xi = [xi lfq.setD.P^(-0.5)*[cos(th(i));sin(th(i))]];
end
lfq.setP.v = xi;
lfq.setP.A = vert2con(xi')';
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(lfq.setP.v'),'alpha',0.35,'color','Orange');
Ellipse_plot(lfq.setD.P,zeros(2,1),[],'Yellow');
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}', '\bf{P}', '\partial\bf{D}_0');
% xlim([-0.2 0.2]), ylim([-11 11]);
xlim([-1.4 1.4]), ylim([-11 11]);

% scale initial proposed DOA such that set D is inside set P
gamma = sdpvar(1);
cons = [];
for i = 1:size(lfq.setP.A,2)
    cons = [cons;[gamma lfq.setP.A(:,i)';lfq.setP.A(:,i) lfq.setD.P] >= 0];
end
t(3) = t(3) + toc;
options = sdpsettings('solver','mosek','verbose',0);
z = optimize(cons,gamma,options);
t(3) = t(3) + z.solvertime;
lfq.setD.P = lfq.setD.P*value(gamma);

figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(lfq.setP.v'),'alpha',0.35,'color','Orange');
Ellipse_plot(lfq.setD.P,zeros(2,1),[],'Yellow');
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}', '\bf{P}', '\partial\bf{D}_0');
% xlim([-0.2 0.2]), ylim([-11 11]);
xlim([-1.4 1.4]), ylim([-11 11]);

% apply algorithm in order to find feedback gain and DOA
lfq.delta = 0.01;
lfq.imax = 20;
[lfq.K,lfq.setP,t0] = doa_quad_model_lfq(mod,lfq.setD.P,...
    lfq.setP,setU,lfq.delta,lfq.imax);
t(3) = t(3) + t0;
tic;
gamma = 0;
for i = 1:size(lfq.setP.A,2)
    gamma = max(gamma,lfq.setP.A(:,i)'*lfq.setD.P^(-1)*lfq.setP.A(:,i));
end
lfq.setD.P = lfq.setD.P*gamma;
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(lfq.setP.v'),'alpha',0.35,'color','Orange');
Ellipse_plot(lfq.setD.P,zeros(2,1),[],'Yellow');
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}', '\bf{P}', '\partial\bf{D}');
% xlim([-0.07 0.07]), ylim([-3.1 3.1]);
xlim([-1.4 1.4]), ylim([-11 11]);

t(3) = t(3) + toc;
vol(3) = pi*prod(1./sqrt(eig(lfq.setD.P)))/volume(Polyhedron(setX.v'));

%% Nonlinear feedback using quadratic Lyapunov function
% model
qlq.mod.A0 = mod.A0;
qlq.mod.B0 = mod.B0;
for i = 1:size(mod.B0,1)
    qlq.mod.B{i} = zeros(size(mod.B0));
    for j = 1:size(mod.B0,1)
        qlq.mod.B{i}(j,:) = mod.B{j}(i,:);
    end
end
% if (Vo < Vin)
%     qlq.setP.v = [0 0;0 2*mod.xe(2);2*mod.xe';2*mod.xe(1) 0]' - mod.xe;
% else
%     qlq.setP.v = [2 50;2 2*mod.xe(2)-50;(2*mod.xe-[2;50])';2*mod.xe(1)-2 50]' - mod.xe;
% end
% qlq.setP.A = vert2con(qlq.setP.v')';
qlq.setP = setX;
[qlq.P,qlq.F,t0] = doa_quad_model_qfq(qlq.mod,qlq.setP,setU);
% xlim([-0.2 0.2]), ylim([-6.5 6.5]);
xlim([-1.4 1.4]), ylim([-15 15]);
xlabel('x^{(1)}'), ylabel('x^{(2)}');
% legend('\bf{X}','\hat{\bf{P}}','\partial\bf{D}');
set(legend( '$$\mathbf{X}$$', '$$\hat{\mathbf{P}}$$', '$$\partial\mathbf{D}$$'), 'Interpreter','Latex','FontSize', 10);
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, Ellipse_plot(qlq.P,zeros(2,1),[],'Yellow');
% xlim([-0.2 0.2]), ylim([-6.5 6.5]);
xlim([-0.9 0.9]), ylim([-15 15]);
legend('\bf{X}','\partial\bf{D}');
xlabel('x^{(1)}'), ylabel('x^{(2)}');

t(4) = t0;
vol(4) = pi*prod(1./sqrt(eig(qlq.P)))/volume(Polyhedron(setX.v'));

%%
X = categorical({'random sampling','lin. Lyap.','quad. Lyap., lin. contr.',...
    'quad. Lyap., nonlin. contr.'});
figure, bar(X,[vol/max(vol);t/max(t)]');
grid;
legend('relative volume','relative time','location','best');
ylabel('relative values')