clear, close all;
Vin = 240;
p = 6;
Rs = 1.5;
lambda_pm = 0.05;
Lq = 1.1e-4;
Ld = 3.5e-4;
JL = 0.019;
DL = 0.12;
TL0 = 30;
Ts = 1e-4;
omega0 = 140;

% equilibrium point of the pmsm
mod.xe(1,1) = 0;
mod.xe(2,1) = 2*(DL*omega0 + TL0)/(sqrt(3)*p*lambda_pm);
mod.xe(3,1) = omega0;
mod.ue(1,1) = -p*Lq*omega0*mod.xe(2)/3;
mod.ue(2,1) = (Rs*mod.xe(2) + sqrt(3)*p*lambda_pm*omega0)/3;

% model pmsm
mod.A0 = [1-Rs*Ts/Ld 0 0;0 1-Rs*Ts/Lq -sqrt(3)*p*lambda_pm*Ts/Lq;...
    0 sqrt(3)*p*lambda_pm*Ts/(2*JL) 1-DL*Ts/JL];
mod.A{1} = zeros(3);
mod.A{1}(3,2) = p*Lq*Ts/Ld;
mod.A{2} = zeros(3);
mod.A{2}(3,1) = -p*Ld*Ts/Lq;
mod.A{3} = zeros(3);
mod.A{3}(1,2) = p*(Ld-Lq)*Ts/(2*JL);
mod.B0 = [3*Ts/Ld 0;0 3*Ts/Lq;0 0];

% shift model such that the origin is the equilibrium point
mod.A0 = mod.A0 + [mod.xe'*mod.A{1};mod.xe'*mod.A{2};mod.xe'*mod.A{3}] + ...
    [mod.xe'*mod.A{1}';mod.xe'*mod.A{2}';mod.xe'*mod.A{3}'];

% set of allowable states
X = kron([-1 1],[210;210;150]);
X(3,1) = 0;
setX.v = interval2vert(X)' - mod.xe;
setX.A = vert2con(setX.v')';

% set of allowable inputs
th = linspace(0,2*pi,9);
setU.v = (Vin/sqrt(3))*[cos(th(1:end-1));sin(th(1:end-1))] - mod.ue;
[setU.M, setU.um] = vert2con(setU.v');

%% random sampling approach
% linear quadratic controller
tic;
rsa.Q = blkdiag(1,1,0.001);
rsa.R = blkdiag(1,1);
[rsa.P0,~,rsa.K] = dare(mod.A0,mod.B0,rsa.Q,rsa.R);
rsa.K = -rsa.K;

% determine an estimation of the DOA;
modelfun = @(x)PMSMModel(x,mod,rsa);
rsa.delta = 0.001;
rsa.past = 0.999;
rsa.mu = 0.001;
rsa.tau = 0.001;
t(1) = toc;
[rsa.P0,t0] = random_sampling_doa(setX,setU,rsa.P0,modelfun,rsa.delta,...
    rsa.past,rsa.mu,rsa.tau);
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
t(1) = t(1) + t0;
vol(1) = pi*prod(1./sqrt(eig(rsa.P0)))/volume(Polyhedron(setX.v'));

%% Linear feedback using linear Lyapunov function
% check whether setX is sufficient for finding a stabilising controller
[K,epsilon,t0] = LFLDetContr2(mod,setX,setU);
t(2) = t0;
% pole placement
tic;
% lfl.lambda = [0.8,0.85,0.999];
lfl.lambda = [0.8,0.85,0.9];
lfl.K0 = -place(mod.A0,mod.B0,lfl.lambda);

% determine initial polytope (set P)
[P,~] = eig(mod.A0 + mod.B0*lfl.K0);
lfl.setP.A = (kron(blkdiag(1,1,1),[1;-1])*P^(-1))';
lfl.setP.v = con2vert(lfl.setP.A')';

figure, plot(Polyhedron(lfl.setP.v'),'alpha',0.35,'color','Orange');
hold on, plot3(0,0,0,'r*');
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');

% scale initial polytope (set P)
alpha = sdpvar(1);
cons = [];
for i = 1:size(lfl.setP.v,2)
    cons = [cons;setX.A'*lfl.setP.v(:,i)*alpha <= ones(size(setX.A,2),1)];
    cons = [cons;setU.M*lfl.K0*lfl.setP.v(:,i)*alpha <= setU.um];
end
t(2) = t(2) + toc;
options = sdpsettings('solver','linprog','verbose',0);
z = optimize(cons,-alpha,options);
t(2) = t(2) + z.solvertime;
lfl.setP.v = lfl.setP.v*value(alpha);
lfl.setP.A = vert2con(lfl.setP.v')';

figure, plot(Polyhedron(lfl.setP.v'),'alpha',0.35,'color','Orange');
hold on, plot3(0,0,0,'r*');
xlim([-20 20]), ylim([-20 20]), zlim([-10 10]);
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');

% determine controller corresponding by initial set P
[lfl.K0,epsilon,t0] = LFLDetContr2(mod,lfl.setP,setU);
t(2) = t(2) + t0;

% run algorithm
lfl.s0 = 1;
lfl.sast = 0.05;
lfl.maxiter = 1000;
lfl.add_vertices = 0;
lfl.zeta = 2;
% lfl.zeta = (sqrt(5)+1)/2;
% lfl.zeta = fibonacci(3:10);
[lfl.K,lfl.setP,t0] = doa_quad_model_lfl(lfl.setP,lfl.s0,lfl.sast,...
    lfl.zeta,lfl.maxiter,lfl.add_vertices,setX,setU,mod);
t(2) = t(2) + t0;
lfl.setP = setP{end};
lfl.K = K{end};
% t(2) = t(2) + 5.172692654250001e+04;
figure, plot(Polyhedron(setX.v'),'alpha',0.1,'color','Blue')
hold on, plot(Polyhedron(lfl.setP.v'),'alpha',1,'color','Orange')
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
plot3(0,0,0,'r*');
legend('\bf{X}', '\bf{P}');

vol(2) = volume(Polyhedron(lfl.setP.v'))/volume(Polyhedron(setX.v'));

%% Linear feedback using quadratic Lyapunov function
% determine initial proposed DOA
tic;
P = sdpvar(3);
Q = sdpvar(3);
R = sdpvar(2);
cons = [];
for i = 1:size(setX.A,2)
    cons = [cons;[1 setX.A(:,i)';setX.A(:,i) P] >= eps];
end
cons = [cons;[mod.A0'*P*mod.A0-P+Q mod.A0'*P*mod.B0;...
    mod.B0'*P*mod.A0 mod.B0'*P*mod.B0+R] >= 0];
cons = [cons;P >= eps;Q >= 0; R >= eps];
obj = trace(P);
t(3) = toc;
options = sdpsettings('solver','mosek','verbose',0);
z = optimize(cons,obj,options);
t(3) = t(3) + z.solvertime;
lfq.setD.P = value(P);

% determine polytope P
tic;
N = 12;
th = linspace(0,pi,N+1);
xi = [];
for i = 1:N
    for j = 1:N
        xi = [xi lfq.setD.P^(-0.5)*[cos(th(i));sin(th(i))*cos(2*th(j));sin(th(i))*sin(2*th(j))]];
    end
end
lfq.setP.v = xi;
lfq.setP.A = vert2con(xi')';

% scale initial proposed DOA such that set D is inside set P
gamma = 0;
for i = 1:size(lfq.setP.A,2)
    gamma = max(gamma,lfq.setP.A(:,i)'*lfq.setD.P^(-1)*lfq.setP.A(:,i));
end
lfq.setD.P = lfq.setD.P*value(gamma);
t(3) = t(3) + toc;
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(lfq.setP.v'),'alpha',0.35,'color','Orange');
Ellipse_plot(lfq.setD.P,zeros(3,1));
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
legend('\bf{X}', '\bf{P}', '\partial\bf{D}_0');
xlim([-220 220]), ylim([-35 35]), zlim([-11 11]);

% apply algorithm in order to find feedback gain and DOA
lfq.delta = 0.01;
lfq.imax = 20;
[lfq.K,lfq.setP,t0] = doa_quad_model_lfq(mod,lfq.setD.P,...
    lfq.setP,setU,lfq.delta,lfq.imax);
t(3) = t(3) + t0;
tic;
gamma = 0;
for i = 1:size(lfq.setP.A,2)
    gamma = max(gamma,lfq.setP.A(:,i)'*lfq.setD.P^(-1)*lfq.setP.A(:,i));
end
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(lfq.setP.v'),'alpha',0.35,'color','Orange');
Ellipse_plot(lfq.setD.P*gamma,zeros(3,1));
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}', '\bf{P}', '\partial\bf{D}');
xlim([-220 220]), ylim([-35 35]), zlim([-11 11]);

t(3) = t(3) + toc;
vol(3) = pi*prod(1./sqrt(eig(lfq.setD.P)))/volume(Polyhedron(setX.v'));

%% Nonlinear feedback using quadratic Lyapunov function
% model
qlq.mod.A0 = mod.A0;
qlq.mod.B0 = mod.B0;
for i = 1:size(mod.A0,1)
    qlq.mod.A{i} = zeros(size(mod.A0));
    for j = 1:size(mod.A0,1)
        qlq.mod.A{i}(j,:) = mod.A{j}(i,:);
    end
end
% qlq.setP.v = [210 210 150;210 210 2*mod.xe(3)-150;210 2*mod.xe(2)-210 2*mod.xe(3)-150;...
%     210 2*mod.xe(2)-210 150;2*mod.xe(1)-210 210 150;2*mod.xe(1)-210 210 2*mod.xe(3)-150;...
%     (2*mod.xe-[210;210;150])';2*mod.xe(1)-210 2*mod.xe(2)-210 150]' - mod.xe;
% qlq.setP.A = vert2con(qlq.setP.v')';
qlq.setP = setX;
[qlq.P,qlq.F,t0] = doa_quad_model_qfq(qlq.mod,qlq.setP,setU);
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
legend('\bf{X}','\bf{P}','\partial\bf{D}');
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, Ellipse_plot(qlq.P,zeros(3,1),[],'Yellow');
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
% legend('\bf{X}','\partial\bf{D}');
xlim([-60 60]), ylim([-12 12]), zlim([-5 5]);

t(4) = t0;
vol(4) = pi*prod(1./sqrt(eig(qlq.P)))/volume(Polyhedron(setX.v'));

%%
X = categorical({'random sampling','lin. Lyap.','quad. Lyap., lin. contr.',...
    'quad. Lyap., nonlin. contr.'});
figure, bar(X,[vol/max(vol);t/t(4)]');
ylim([0 1.2]);
grid;
legend('relative volume','relative time','location','best');
ylabel('relative values');