function [K,setP,t] = doa_quad_model_lfq(mod,P,setP,setU,delta,imax)
% Determine the local controller (linear control gain) with corresponding
% domain of attraction using a quadratic Lyapunov function.
% [K,setP,t] = doa_quad_model_lfq(mod,P,setP,setU,delta,imax)
% 
% inputs:
% mod : quadratic model of the system (A0, A{}, B0, B{})
% P : initial shape of the DOA (level curve of the Lyapunov function)
% setP : structure consisting of at least the vertices (v) of the
% polynomial around the initial proposed DOA
% setU : set of allowable inputs (M*u <= um)
% delta : margin in the decrease of the Lyapunov function
% imax : maximum number of iterations of the algorithm
% 
% outputs:
% K : linear controller gain
% setP : DOA of the local controller
% t : computational time
% 
% Rik Koch (TU Delft, 2021)

n = size(setP.v,1);
m = size(setU.M,2);
p = size(setP.v,2);
t = 0;
i = 1;
while(1)
    rhoi = sdpvar(1);
    K = sdpvar(m,n);
    cons = [];
    for j = 1:p
        A = mod.A0;
        B = mod.B0;
        for k = 1:n
            if(isfield(mod,'A'))
                A(k,:) = A(k,:) + setP.v(:,j)'*mod.A{k};
            end
            if(isfield(mod,'B'))
                B(k,:) = B(k,:) + setP.v(:,j)'*mod.B{k};
            end
        end
        cons = [cons;[rhoi*P (A + B*K)'*P;P*(A + B*K) P] >= eps];
        cons = [cons;setU.M*K*setP.v(:,j) <= setU.um];
    end
    if (i > 1)
        rhop = 1;
        for j = 1:i-1
            rhop = rhop*rho(j);
        end
        cons = [cons;rhop*rhoi >= 1+eps];
        cons = [cons;rhoi <= 1 + abs(1-rho(i-1)) - eps];
    end
    options = sdpsettings('solver','mosek','verbose',0);
    z = optimize(cons,rhoi,options);
    t = t + z.solvertime;
    if (z.info(1) == 'S')
        rho(i) = value(rhoi);
    else
        error('optimisation problem not solvable');
    end
    if (rho(i) > 1)
        if (rho(i) < 1+delta)
            setP.v = 0.99*setP.v;
        end
        setP.v = setP.v/rho(i);
    elseif (rho(i) <= 1)
        if (i == 1)
            break;
        end
        if (rho(i) < 1-delta && i < imax)
            setP.v = setP.v/rho(i);
        else
            setP.v = setP.v/rho(i);
            break;
        end
    end
    i = i + 1;
end
figure, plot(rho,'b-o');
hold on, plot(ones(max(size(rho))+1,1),'r--');
grid;
xlabel('i'), ylabel('\rho_i');
setP.A = vert2con(setP.v')';
K = value(K);
end