function [P,Fhat,t] = doa_quad_model_qfq(mod,setP,setU)
% This function determines the nonlinear controller based on a quadratic
% Lyapunov function with corresponding domain of attraction.
% [P,Fhat,t] = doa_quad_model_qfq(mod,setP,setU)
% 
% inputs:
% mod : quadratic model of the system (A0, A{}, B0, B{})
% setP : polynomial set (A*x <= 1, v are the vertices) which will contain
% the obtained DOA computed using this method (usually equal to setX)
% setU : set of allowable inputs (M*u <= um)
% 
% outputs:
% P : DOA of the local controller (x'*P*x <= 1)
% Fhat : nonlinear feedback matrices of the local controller (see report
% for details)
% t : computational time
% 
% Rik Koch (TU Delft, 2021)

% define parameters
n = size(mod.A0,1);
m = size(mod.B0,2);
p = size(setP.A,2);
q = size(setP.v,2);
v = setP.v;
t = 0;
tic;
% Bhat
Bhat = zeros(n^2,m);
if (isfield(mod,'B'))
    for i = 1:n
        Bhat((i-1)*n+1:i*n) = mod.B{i};
    end
end

% N1 and N2
N1 = [eye(n) zeros(n,n^2)];
N2 = [zeros(n^2,n) eye(n^2)];

for i = 1:n+1
    Fhat{i} = sdpvar(m,n);
end
Fhat{end+1} = sdpvar(n*m,n^2);
% Fhat{end+1} = zeros(n*m,n^2);
L = sdpvar(2*n*(n+1),2*(2*n^2-n));
Q = sdpvar(n);
W = sdpvar(n);
Z = sdpvar(n*(n+1));

% optimisation problem
cons = [];
cons = [cons;[W eye(n);eye(n) Q] >= 0];
cons = [cons;Q >= eps];
for i = 1:p
    cons = [cons;setP.A(:,i)'*Q*setP.A(:,i) <= 1];
end
for i = 1:q
    % Pi and Piu
    Pi = kron(v(:,i),eye(n));
    Piu = kron(v(:,i),eye(m));
    
    % A
    A = mod.A0;
    if (isfield(mod,'A'))
        for j = 1:n
            A = A + v(j,i)*mod.A{j};
        end
    end
    
    F1 = Fhat{1};
    for j = 1:n
        F1 = F1 + v(j,i)*Fhat{j+1};
    end
    
    % Psi
    Psi = [eye(n);Pi];
    
    % Lambda
    Lambda = kron([diag(v(2:end,i)) zeros(n-1,1)] + [zeros(n-1,1) -diag(v(1:end-1,i))], eye(n));
    
    % Omegahat
    Omegahat = [kron(eye(2),[Pi -eye(n^2)]);kron(eye(2),[zeros(n*(n-1),n) Lambda])];
    
    % Phi
    Phi(1:n*(n+1),1:n*(n+1)) = -N1'*Q*N1;
    Phi(n*(n+1)+1:2*n*(n+1),1:n*(n+1)) = N1'*(A*Q + mod.B0*F1)*N1 + ...
        N2'*Bhat*F1*N1 + N1'*mod.B0*Piu'*Fhat{end}*N2 + ...
        N2'*Bhat*Piu'*Fhat{end}*N2;
    Phi(1:n*(n+1),n*(n+1)+1:2*n*(n+1)) = Phi(n*(n+1)+1:2*n*(n+1),1:n*(n+1))';
    Phi(n*(n+1)+1:2*n*(n+1),n*(n+1)+1:2*n*(n+1)) = -(Z + Z');
    
    % Gamma
    Gamma = [zeros(n,n*(n+1)) Psi'*Z];
    
    cons = [cons;[Phi+L*Omegahat+Omegahat'*L' Gamma';Gamma -Q] <= eps];
end

obj = trace(W);
t = t + toc;
options = sdpsettings('solver','mosek','verbose',0);
z = optimize(cons,obj,options);
t = t + z.solvertime;
if (z.info(1:19) == 'Successfully solved')
    P = value(Q)^(-1);
    for i = 1:n+2
        F{i} = value(Fhat{i});
    end
    Fhat = F;
else
    warning(z.info);
    P = [];
    Fhat = [];
    return;
end

% scale set based on input
bounds = findboundsellipsoid(P);
vhat = interval2vert(bounds)';
figure, plot(Polyhedron(setP.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(vhat'),'alpha',0.35,'color','Orange');
Ellipse_plot(P,zeros(n,1),[],'Yellow');
gamma = sdpvar(1);
cons = [];
cons = [cons; gamma >= 0; gamma <= 1];
T1 = Fhat{1}*P;
for j = 1:q
    T2 = zeros(m,n);
    for i = 1:n
        T2 = T2 + vhat(i,j)*Fhat{i+1}*P;
    end
    for k = 1:q
        T3 = kron(vhat(:,j),eye(m))'*Fhat{end}*kron(vhat(:,k),eye(n))*P;
        for l = 1:q
            cons = [cons;gamma*setU.M*(T1 + T2 + T3)*vhat(:,l) <= setU.um];
        end
    end
end
options = sdpsettings('solver','linprog','verbose',0);
z = optimize(cons,-gamma,options);
t = t + z.solvertime;
if (z.info(1) == 'S')
    gamma = value(gamma);
else
    return;
end

disp(gamma)
setPhat.A = vert2con(gamma*vhat')';
gamma = 0;
for i = 1:p
    gamma = max(gamma,setPhat.A(:,i)'*value(Q)*setPhat.A(:,i));
end
disp(gamma);
P = P*gamma;
figure, plot(Polyhedron(setP.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(vhat'),'alpha',0.35,'color','Orange');
Ellipse_plot(P,zeros(n,1),[],'Yellow');
end