function [P0,t0] = random_sampling_doa(setX,setU,P0,modelfun,delta,past,mu,tau)
% This function determines the local controller (linear) with corresponding
% domain of attraction using a quadratic Lyapunov function using samples of
% the state space.
% [P0,t0] = random_sampling_doa(setX,setU,P0,modelfun,delta,past,mu,tau)
% 
% inputs:
% setX : set of allowable states (A*x <= 1)
% setU : set of allowable inputs (M*x <= um)
% P0 : initial size of the DOA (x'*P0*x <= 1)
% modelfun : function of the quadratic model
% delta : margin in the probability of getting a reliable DOA
% past : margin in the probability of getting a reliable DOA
% mu : maring with respect to the boundary of setX
% tau : increasing reliability of the obtained DOA
% 
% outputs:
% P0 : DOA of the local controller (x'*P0*x <= 1)
% t0 : computational time
% 
% Rik Koch (TU Delft, 2021)

tic;
% Scale initial set
t = max(diag(setX.A'*P0^(-1)*setX.A));
P0 = P0*t;

% Set number of samples
Ns = ceil(log(delta)/log(past));
t0 = toc;

% iterate over proposed DOA's
while(1)
    tic;
    vmin = 1; % maximum value of the Lyapunov function is 1
    x = random_sampling_ellipse(P0,Ns,"ellipse"); % sampling using Halton sequence
    v = diag(x'*P0*x); % Lyapunov function for x
    [xp,u] = modelfun(x); % calculation of next state and input
    vp = diag(xp'*P0*xp); % Lyapunov function for xp
    dv = vp - v; % Lyapunov's second criterion
    iu = (dv > 0) | max(setU.M*u > setU.um,[],1)';
    Us = x(:,iu); % set consists of states that do not satisfy conditions
    ip = not(iu) & (abs(max(setX.A'*x - 1,[],1)) <= mu)' | max(setX.A'*x - 1 > 0,[],1)';
    Ps = x(:,ip); % set consists of states that are too close to the border of setX
    ij = not(iu) & not(ip);
    Js = x(:,ij); % set consists of states that form the DOA
    t0 = t0 + toc;
    if (size(P0,1) <= 3)
        figure, Ellipse_plot(P0,zeros(size(P0,1),1));
        grid on, hold on;
        if (~isempty(Us))
            if (size(P0,1) == 2)
                plot(Us(1,:),Us(2,:),'r.');
            elseif (size(P0,1) == 3)
                plot3(Us(1,:),Us(2,:),Us(3,:),'r.');
            end
        end
        if (~isempty(Ps))
            if (size(P0,1) == 2)
                plot(Ps(1,:),Ps(2,:),'b.');
            elseif (size(P0,1) == 3)
                plot3(Ps(1,:),Ps(2,:),Ps(3,:),'b.');
            end
        end
        if (~isempty(Js))
            if (size(P0,1) == 2)
                plot(Js(1,:),Js(2,:),'g.');
            elseif (size(P0,1) == 3)
                plot3(Js(1,:),Js(2,:),Js(3,:),'g.');
            end
        end
        title('Samples drawn from the ellipse');
        xlabel('x^{(1)}'), ylabel('x^{(2)}');
        if (size(P0,1) == 3)
            zlabel('x^{(3)}');
        end
    end
    tic;
    vmin = min(vmin,min(v(iu | ip)));
    if (sum(iu) == 0)
        if (sum(ip) > 0)
            P0 = P0/vmin;
        end
        break;
    else
        P0 = P0/(vmin-tau); % new proposed DOA
    end
    t0 = t0 + toc;
end
end