function x = random_sampling_ellipse(P,Ns,method)
% Drawing samples inside an ellipsoid using the halton sequence.
% x = random_sampling_ellipse(P,Ns,method)
% 
% inputs:
% P : shape of the ellipsoid (x'*P*x = 1 is the boundary of the ellipsoid)
% Ns : number of samples
% method : applied method (either ellips or simplex) (see the report for
% more information)
% 
% outputs:
% x : quasi-random samples (drawn using halton sequence)
% 
% Rik Koch (TU Delft, 2021)

n = size(P,1);
if(~issymmetric(P))
    error('matrix P is not symmetric');
end
if (method == "ellipse")
    q = qrandstream('halton',n,'Skip',1e3,'Leap',1e2);
    qb = qrandstream('halton',n-1,'Skip',1e5,'Leap',1e2);
    zeta = [qrand(q,Ns);ones(ceil(Ns/10),1) qrand(qb,ceil(Ns/10))];
    Ns = size(zeta,1);
    r = nthroot(zeta(:,1),n)';
    th = zeros(Ns,n-1);
    if (n > 2)
        if (n > 3)
            load('pwainvellipse.mat');
            a = pwainvellipse.a;
            b = pwainvellipse.b;
            c = pwainvellipse.c;
            for i = 1:n-3
                for j = 1:Ns
                    for k = 1:size(c,1)-1
                        if(zeta(j,i+1) <= c(k+1))
                            th(j,i) = a(k,n-1-i)*zeta(j,i+1) + b(k,n-1-i);
                        end
                    end
                end
            end
        end
        th(:,n-2) = acos(1 - 2.*zeta(:,n-1));
    end
    th(:,n-1) = 2*pi*zeta(:,n);
    [V,D] = eig(P);
    x = diag(D^(-0.5))*r;
    for i = 1:n-1
        x(i,:) = x(i,:).*cos(th(:,i)');
        x(i+1:n,:) = x(i+1:n,:).*sin(th(:,i)');
    end
    x = V*x;
elseif(method == "simplex")
    % determine vertices of polytope (points on the ellipse)
    N = 10;
    th1 = linspace(0,pi*(N-1)/N,N);
    th2 = linspace(0,2*pi*(N-1)/N,N);
    [V,D] = eig(P);
    k = zeros(n-1,N^(n-1));
    for i = 1:n-1
        k(i,:) = repmat(kron(1:N,ones(1,N^(i-1))),1,N^(n-1-i));
    end
    xi = repmat(diag(D^(-0.5)),1,N^(n-1));
    for j = 1:N^(n-1)
        for i = 1:n-2
            xi(i,j) = xi(i,j)*cos(th1(k(i,j)));
            xi(i+1:n,j) = xi(i+1:n,j).*sin(th1(k(i,j)));
        end
        xi(n-1,j) = xi(n-1,j)*cos(th2(k(n-1,j)));
        xi(n,j) = xi(n,j)*sin(th2(k(n-1,j)));
    end
    xi = V*xi;
    
    % scale polytope such that the ellipse is fully inside the polytope
    A = vert2con(xi');
    t = max(diag(A*P^(-1)*A'));
    xi = con2vert(A)*nthroot(t,n);
    
    % draw random samples in the polytope
    q = qrandstream('halton',n+1,'Skip',1e3,'Leap',1e2);
    qb = qrandstream('halton',n,'Skip',1e5,'Leap',1e2);
    zeta = [qrand(q,Ns);qrand(qb,ceil(Ns/10)) ones(ceil(Ns/10),1)];
    zeta = zeta(:,[1:n-1 n+1 n]);
    SMP = convhulln(xi); % division of polytope into simplices
    Nsmp = size(SMP,1); % number of simplices
    [~,~,simpind] = histcounts(zeta(:,n+1),cumsum([0;ones(Nsmp,1)/Nsmp])); % determine the chosen simplex
    x = xi(SMP(simpind,1),:);
    for i = 1:n-1
        x = x.*nthroot(zeta(:,i),i) + xi(SMP(simpind,i+1),:).*(1-nthroot(zeta(:,i),i));
    end
    x = x.*nthroot(zeta(:,n),n);
    x = x';
else
    error('Invalid method');
end
end