function uc = BBC_sdnmpc(xc)
% Application of the SD-NMPC method to the system of the buck-boost
% converter in BuckBoostConverterController.slx.
% uc = BBC_sdnmpc(xc)
% xc is the current state and uc the computed next control input
% 
% Rik Koch (TU Delft, 2021)

global mdl
persistent iter
persistent iseq
persistent lcon
persistent sdc

% determination of the local controller with corresponding domain of
% attraction
if (isempty(lcon))
    lfl.lambda = [0.8,0.9];
    lfl.s0 = 1;
    lfl.sast = 0.05;
    lfl.maxiter = 200;
    lfl.add_vertices = 0;
    lfl.zeta = 2;
    [K,epsilon,~] = LFLDetContr2(mdl,mdl.setX,mdl.setU);
    if (epsilon < 1)
        lcon.K = K;
        lcon.setD = setX;
    else
        K0 = -place(mdl.A0,mdl.B0,lfl.lambda);
        [P,~] = eig(mdl.A0 + mdl.B0*K0);
        if (sum(imag(lfl.lambda) == 0) == 0)
            rho = real(lfl.lambda(1));
            beta = abs(imag(lfl.lambda(1)));
            N = ceil(2*pi/beta);
            W = zeros(N,2);
            for k = 1:N
                W(k,:) = [cos((2*k-1)*pi/N) sin((2*k-1)*pi/N)];
            end
            w = cos(pi/N)*ones(N,1);
            P = [real(P(:,1)) imag(P(:,1))];
            setP.A = (W*P^(-1)./w)';
            setP.v = con2vert(setP.A')';
        else
            setP.A = (kron(eye(2),[1;-1])*P^(-1))';
            setP.v = con2vert(setP.A')';
        end
        
        alpha = sdpvar(1);
        cons = [];
        for i = 1:size(setP.v,2)
            cons = [cons;mdl.setX.A'*setP.v(:,i)*alpha <= ones(size(mdl.setX.A,2),1)];
            cons = [cons;mdl.setU.M*K0*setP.v(:,i)*alpha <= mdl.setU.um];
        end
        options = sdpsettings('solver','linprog','verbose',0);
        z = optimize(cons,-alpha,options);
        setP.v = setP.v*value(alpha);
        setP.A = vert2con(setP.v')';
        
        [K0,~,~] = LFLDetContr2(mdl,setP,mdl.setU);
        
        [lcon.K,lcon.setD,~] = doa_quad_model_lfl(setP,lfl.s0,lfl.sast,...
            lfl.zeta,lfl.maxiter,lfl.add_vertices,mdl.setX,mdl.setU,mdl);
    end
end

% Determination of the initial control sequence
if (isempty(iseq))
    lpv.Nhat = 5;
    lpv.p = 1;
    lpv.x0 = mdl.x0 - mdl.xe;
    lpv.P = blkdiag(3,1,1);
    lpv.Q = blkdiag(3,1,1);
    lpv.R = 0.1;
    lpv.N = 1000;
    if (isfield(lcon.setD,'P'))
        [lpv.x,lpv.u,~,~] = initseq_lpv_mpc_v2(mdl,lpv,mdl.setX,mdl.setU,lcon.setD.P);
    elseif (isfield(lcon.setD,'A'))
        [lpv.x,lpv.u,~,~] = initseq_lpv_mpc_v2(mdl,lpv,mdl.setX,mdl.setU,lcon.setD);
    else
        error('lcon not as expected')
    end
    lpv.N = size(lpv.x,2)-1;
    iseq.Uw = lpv.u + mdl.ue;
    iseq.N = lpv.N;
end

% Initialization of the online SD-NMPC algorithm
if (isempty(iter))
    close all;
    iter = 0;
    sdc.Ns = 5;
    sdc.Nc = 5;
    sdc.Q = blkdiag(4,1);
    sdc.R = 1;
    sdc.updates = 0;
    sdc.par.method = 'uniform';
    sdc.par.mu = mdl.ue;
    sdc.par.sigma = 0.1;
    sdc.sampfun = @(s,par)sdnmpc_sampling_BBC(s,par);
    sdc.m = size(mdl.setU.v,1);
    if (isfield(lcon.setD,'A'))
        sdc.ad = 1;
    elseif (isfield(lcon.setD,'P'))
        sdc.ad = 0;
    else
        error('invalid DOA');
    end
    
    if (iseq.N < sdc.Nc)
        x = [xc zeros(size(mdl.xe,1),sdc.Nc)];
        for i = 1:sdc.Nc
            if (i > iseq.N)
                iseq.Uw(:,i) = lcon.K*(x(:,i) - mdl.xe) + mdl.ue;
            end
            x(:,i+1) = mdl.fun(x(:,i) - mdl.xe,iseq.Uw(:,i) - mdl.ue) + mdl.xe;
        end
    end
    iseq.N = size(iseq.Uw,2);
end

% Application of the online SD-NMPC algorithm
iter = iter + 1;
m = size(mdl.ue,1);
n = size(mdl.xe,1);
updates = 0;
tic;
sdc.ncom(iter) = 0;
x = [xc zeros(n,sdc.Nc)];
Jsub = 0;
for i = 1:sdc.Nc
    sdc.ncom(iter) = sdc.ncom(iter) + 2;
    x(:,i+1) = mdl.fun(x(:,i) - mdl.xe,iseq.Uw(:,i) - mdl.ue) + mdl.xe;
    Jsub = Jsub + (x(:,i+1) - mdl.xe)'*sdc.Q*(x(:,i+1) - mdl.xe) + ...
        (iseq.Uw(:,i) - mdl.ue)'*sdc.R*(iseq.Uw(:,i) - mdl.ue);
end
xf = x(:,end);
p = haltonset(m,'Skip',iter*1e3,'Leap',1e2);
s = net(p,sdc.Nc*sdc.Ns);
for i = 1:sdc.Nc
    sdc.par.mu = iseq.Uw(:,i);
    s((i-1)*sdc.Ns+1:i*sdc.Ns,:) = sdc.sampfun(s((i-1)*sdc.Ns+1:i*sdc.Ns,:),sdc.par);
    Uq = iseq.Uw;
    for j = 1:sdc.Ns
        Uq(:,i) = s((i-1)*sdc.Ns+j,:)';
        x = [xc zeros(n,max(iseq.N-iter+1,sdc.Nc))];
        Jnew = 0;
        for l = 1:max(iseq.N-iter+1,sdc.Nc)
            sdc.ncom(iter) = sdc.ncom(iter) + 1;
            x(:,l+1) = mdl.fun(x(:,l) - mdl.xe,Uq(:,l) - mdl.ue) + mdl.xe;
            if (max(mdl.setX.A'*(x(:,l+1) - mdl.xe)) > 1)
                Jnew = inf;
                break;
            end
            if (l <= sdc.Nc)
                sdc.ncom(iter) = sdc.ncom(iter) + 1;
                Jnew = Jnew + (x(:,l+1) - mdl.xe)'*sdc.Q*(x(:,l+1) - mdl.xe) + ...
                    (Uq(:,l) - mdl.ue)'*sdc.R*(Uq(:,l) - mdl.ue);
            end
        end
        if ((sdc.ad == 1 && max(lcon.setD.A'*(x(:,l+1) - mdl.xe)) <= 1) || ...
                (sdc.ad == 0 && (x(:,l+1) - mdl.xe)'*lcon.setD.P*(x(:,l+1) - mdl.xe) <= 1))
            if (Jnew < Jsub)
                updates = updates + 1;
                iseq.Uw = Uq;
                Jsub = Jnew;
                xf = x(:,end);
            end
        end
    end
end
uc = iseq.Uw(:,1);
if (size(iseq.Uw,2) > sdc.Nc)
    iseq.Uw = iseq.Uw(:,2:end);
else
    iseq.Uw = [iseq.Uw(:,2:end) lcon.K*(xf - mdl.xe) + mdl.ue];
end
sdc.J(iter) = Jsub;
sdc.t(iter) = toc;
disp(['iteration: ',num2str(iter),', J = ',num2str(sdc.J(iter)),' (',num2str(updates),' updates)']);
sdc.updates = sdc.updates + updates;
if (sdc.J(iter) < 70.8499)
    iter = iter;
end
end