function uc = PMSM_sdnmpc(xc)
% Application of the SD-NMPC method to the system of the three-phase VSI
% connected to a PMSM converter in PMSMController.slx.
% uc = PMSM_sdnmpc(xc)
% xc is the current state and uc the computed next control input
% 
% Rik Koch (TU Delft, 2021)

global mdl
persistent iter
persistent iseq
persistent lcon
persistent sdc
persistent omega0
persistent omegabnd
persistent iterref
persistent newequi
persistent obj

% determination of the local controller with corresponding domain of
% attraction
if (isempty(iter))
    iterref = 0;
    omegaseq = [75 100 115 125 135;
                50 80 100 120 132];
%     omegaseq = [140;140];
    while(1)
        if (iterref == 0)
            omega0 = mdl.xe(3);
            iterref = 1;
            if (omega0 < omegaseq(1,iterref) + 7)
                break;
            end
        elseif (iterref < size(omegaseq,2) && mdl.xe(3) < omegaseq(1,iterref) + 7 - iterref)
            omega0 = [omegaseq(1,1:iterref)';omega0];
            omegabnd = omegaseq(2,1:iterref)';
            iterref = iterref + 1;
            break;
        elseif (iterref == size(omegaseq,2))
            omega0 = [omegaseq(1,1:iterref)';omega0];
            omegabnd = omegaseq(2,1:iterref)';
            iterref = iterref + 1;
            break;
        else
            iterref = iterref + 1;
        end
    end
end

% Determination of the initial control sequence
if (isempty(lcon))
    for k = 1:iterref
        det_model_PMSM(omega0(k));
        lfq.delta = 0.01;
        lfq.imax = 20;
        P = sdpvar(max(size(mdl.xe)));
        Q = sdpvar(max(size(mdl.xe)));
        R = sdpvar(max(size(mdl.ue)));
        cons = [];
        for i = 1:size(mdl.setX.A,2)
            cons = [cons;[1 mdl.setX.A(:,i)';mdl.setX.A(:,i) P] >= eps];
        end
        cons = [cons;[mdl.A0'*P*mdl.A0-P+Q mdl.A0'*P*mdl.B0;...
            mdl.B0'*P*mdl.A0 mdl.B0'*P*mdl.B0+R] >= 0];
        cons = [cons;P >= eps;Q >= 0; R >= eps];
        obj = trace(P);
        options = sdpsettings('solver','mosek','verbose',0);
        z = optimize(cons,obj,options);
        setD.P = value(P);
        
        % determine polytope P
        N = 12;
        th = linspace(0,pi,N+1);
        xi = [];
        for i = 1:N
            for j = 1:N
                xi = [xi setD.P^(-0.5)*[cos(th(i));sin(th(i))*cos(2*th(j));sin(th(i))*sin(2*th(j))]];
            end
        end
        setP.v = xi;
        setP.A = vert2con(xi')';
        
        gamma = 0;
        for i = 1:size(setP.A,2)
            gamma = max(gamma,setP.A(:,i)'*setD.P^(-1)*setP.A(:,i));
        end
        setD.P = setD.P*value(gamma);
        
        % apply algorithm in order to find feedback gain and DOA
        [lcon.K{k},setP,~] = doa_quad_model_lfq(mdl,setD.P,...
            setP,mdl.setU,lfq.delta,lfq.imax);
        gamma = 0;
        for i = 1:size(setP.A,2)
            gamma = max(gamma,setP.A(:,i)'*setD.P^(-1)*setP.A(:,i));
        end
        lcon.setD.P{k} = setD.P*gamma;
    end
    iterref = 1;
    newequi = 1;
    det_model_PMSM(omega0(iterref));
end

% Initialization of the online SD-NMPC algorithm
if (iterref < max(size(omega0)) && xc(3) > omegabnd(iterref))
    iterref = iterref + 1;
    newequi = 1;
    det_model_PMSM(omega0(iterref));
    mdl.x0 = xc;
    iseq = [];
end

% Application of the online SD-NMPC algorithm
if (isempty(iseq))
    sln.x0 = mdl.x0 - mdl.xe;
    sln.n = size(mdl.A0,1);
    sln.m = size(mdl.B0,2);
    sln.N = 2e4;
    sln.Nc = 5;
    sln.x(:,1) = sln.x0;
    sln.Q = blkdiag(1,1,10);
    sln.setX = mdl.setX;
    sln.setU = mdl.setU;
    sln.modelfun = mdl.fun;
    [~,iseq.Uw,iseq.N,~] = initseq_sln(sln,mdl,lcon.setD.P{iterref},1);
    iseq.Uw = iseq.Uw + mdl.ue;
end

if (isempty(iter))
    close all;
    iter = 0;
    sdc.Ns = 0;
    sdc.Nc = 5;
    sdc.Q = blkdiag(1,1,10);
    sdc.R = blkdiag(1,1);
    sdc.updates = 0;
    sdc.par.method = 'truncated normal';
    sdc.par.mu = mdl.ue;
    sdc.par.sigma = [0.1;0.01];
    sdc.sampfun = @(s,par)sdnmpc_sampling_PMSM(s,par,mdl.Vin);
    sdc.m = size(mdl.setU.v,1);
    if (isfield(lcon.setD,'A'))
        sdc.ad = 1;
    elseif (isfield(lcon.setD,'P'))
        sdc.ad = 0;
    else
        error('invalid DOA');
    end
end

if (newequi)
    newequi = 0;
    if (iseq.N < sdc.Nc)
        x = [xc zeros(size(mdl.xe,1),sdc.Nc)];
        for i = 1:sdc.Nc
            if (i > iseq.N)
                iseq.Uw(:,i) = lcon.K{iterref}*(x(:,i) - mdl.xe) + mdl.ue;
            end
            x(:,i+1) = mdl.fun(x(:,i) - mdl.xe,iseq.Uw(:,i) - mdl.ue) + mdl.xe;
        end
    end
    iseq.N = size(iseq.Uw,2);
end

iter = iter + 1;

m = size(mdl.ue,1);
n = size(mdl.xe,1);
updates = 0;
tic;
sdc.ncom(iter) = 0;
x = [xc zeros(n,sdc.Nc)];
Jsub = 0;
for i = 1:sdc.Nc
    sdc.ncom(iter) = sdc.ncom(iter) + 2;
    x(:,i+1) = mdl.fun(x(:,i) - mdl.xe,iseq.Uw(:,i) - mdl.ue) + mdl.xe;
    Jsub = Jsub + (x(:,i+1) - mdl.xf)'*sdc.Q*(x(:,i+1) - mdl.xf) + ...
        (iseq.Uw(:,i) - mdl.uf)'*sdc.R*(iseq.Uw(:,i) - mdl.uf);
end
xf = x(:,end);
p = haltonset(m,'Skip',iter*1e3,'Leap',1e2);
s = net(p,sdc.Nc*sdc.Ns);
for i = 1:sdc.Nc
    sdc.par.mu = iseq.Uw(:,i);
    s((i-1)*sdc.Ns+1:i*sdc.Ns,:) = sdc.sampfun(s((i-1)*sdc.Ns+1:i*sdc.Ns,:),sdc.par);
    Uq = iseq.Uw;
    for j = 1:sdc.Ns
        Uq(:,i) = s((i-1)*sdc.Ns+j,:)';
        x = [xc zeros(n,max(iseq.N-iter+1,sdc.Nc))];
        Jnew = 0;
        for l = 1:max(iseq.N-iter+1,sdc.Nc)
            sdc.ncom(iter) = sdc.ncom(iter) + 1;
            x(:,l+1) = mdl.fun(x(:,l) - mdl.xe,Uq(:,l) - mdl.ue) + mdl.xe;
            if (max(mdl.setX.A'*(x(:,l+1) - mdl.xe)) > 1)
                Jnew = inf;
                break;
            end
            if (l <= sdc.Nc)
                sdc.ncom(iter) = sdc.ncom(iter) + 1;
                Jnew = Jnew + (x(:,l+1) - mdl.xf)'*sdc.Q*(x(:,l+1) - mdl.xf) + ...
                    (Uq(:,l) - mdl.uf)'*sdc.R*(Uq(:,l) - mdl.uf);
            end
        end
        if ((sdc.ad == 1 && max(lcon.setD.A{iterref}'*(x(:,l+1) - mdl.xe)) <= 1) || ...
                (sdc.ad == 0 && (x(:,l+1) - mdl.xe)'*lcon.setD.P{iterref}*(x(:,l+1) - mdl.xe) <= 1))
            if (Jnew < Jsub)
                updates = updates + 1;
                iseq.Uw = Uq;
                Jsub = Jnew;
                xf = x(:,end);
            end
        end
    end
end
uc = iseq.Uw(:,1);
if (size(iseq.Uw,2) > sdc.Nc)
    iseq.Uw = iseq.Uw(:,2:end);
else
    iseq.Uw = [iseq.Uw(:,2:end) lcon.K{iterref}*(xf - mdl.xe) + mdl.ue];
end
sdc.J(iter) = Jsub;
sdc.t(iter) = toc;
disp(['iteration: ',num2str(iter),', J = ',num2str(sdc.J(iter)),' (',num2str(updates),' updates)']);
sdc.updates = sdc.updates + updates;
if (iter == 15001)
   disp('hoi')
end
end