function sdc = sdnmpc(sdc,xe,ue)
% This function applies the online SD-NMPC controller to the (nonlinear)
% system.
% sdc = sdnmpc(sdc,xe,ue)
% 
% inputs:
% sdc : structure consisting of parameters needed to apply the SD-NMPC
% method (kmax (maximum number of iterations), N (size of the initial
% control sequence), Ns (number of samples per step in the control
% horizon), Nc (control horizon), K (local (linear) feedback gain), setD
% (structure consisting of the DOA corresponding to the local controller
% (A*x <= 1 or x'*P*x <= 1)), Q (weigthing matrix on the states of the
% system), R (weighting matrix on the inputs of the system), x0 (initial
% state), u (current control input sequence), modelfun ((nonlinear) model
% of the system), setX (set of allowable states (A*x <= 1)), setU (set of
% allowable inputs (M*u <= um), par.method (distribution used for sampling
% (either uniform, trapezoidal or truncated normal)), par.mu (mean of
% distribution), par.sigma (standard deviation of distribution), sampfun
% (function applied for sampling))
% xe : equilibrium point (state)
% ue : equilibrium point (input)
% 
% output:
% sdc : structure which also consists of states, inputs and value of the
% objective function (J) computed using this method
% 
% Rik Koch (TU Delft, 2021)

if (isfield(sdc.setD,'A'))
    ad = 1;
elseif (isfield(sdc.setD,'P'))
    ad = 0;
else
    error('invalid DOA');
end
m = size(ue,1);
n = size(xe,1);
sdc.updates = 0;
Uw = sdc.u;
sdc.u = zeros(m,sdc.kmax);
if (sdc.N < sdc.Nc)
    x = [sdc.x0 zeros(n,sdc.Nc)];
    for i = 1:sdc.Nc
        if (i > sdc.N)
            Uw(:,i) = sdc.K*(x(:,i) - xe) + ue;
        end
        x(:,i+1) = sdc.modelfun(x(:,i) - xe,Uw(:,i) - ue) + xe;
    end
end
sdc.x = [sdc.x0 zeros(n,sdc.kmax)];
k = 1;
while(k <= sdc.kmax)
    updates = 0;
    tic;
    sdc.ncom(k) = 0;
    x = [sdc.x(:,k) zeros(n,sdc.Nc)];
    Jsub = 0;
    for i = 1:sdc.Nc
        sdc.ncom(k) = sdc.ncom(k) + 2;
        x(:,i+1) = sdc.modelfun(x(:,i) - xe,Uw(:,i) - ue) + xe;
        Jsub = Jsub + (x(:,i+1) - xe)'*sdc.Q*(x(:,i+1) - xe) + ...
            (Uw(:,i) - ue)'*sdc.R*(Uw(:,i) - ue);
    end
    xf = x(:,end);
    p = haltonset(m,'Skip',k*1e3,'Leap',1e2);
    s = net(p,sdc.Nc*sdc.Ns);
    for i = 1:sdc.Nc
        sdc.par.mu = Uw(:,i);
        s((i-1)*sdc.Ns+1:i*sdc.Ns,:) = sdc.sampfun(s((i-1)*sdc.Ns+1:i*sdc.Ns,:),sdc.par);
        Uq = Uw;
        for j = 1:sdc.Ns
            Uq(:,i) = s((i-1)*sdc.Ns+j,:)';
            x = [sdc.x(:,k) zeros(n,max(sdc.N-k+1,sdc.Nc))];
            Jnew = 0;
            for l = 1:max(sdc.N-k+1,sdc.Nc)
                sdc.ncom(k) = sdc.ncom(k) + 1;
                x(:,l+1) = sdc.modelfun(x(:,l) - xe,Uq(:,l) - ue) + xe;
                if (max(sdc.setX.A'*(x(:,l+1) - xe)) > 1)
                    Jnew = inf;
                    break;
                end
                if (l <= sdc.Nc)
                    sdc.ncom(k) = sdc.ncom(k) + 1;
                    Jnew = Jnew + (x(:,l+1) - xe)'*sdc.Q*(x(:,l+1) - xe) + ...
                        (Uq(:,l) - ue)'*sdc.R*(Uq(:,l) - ue);
                end
            end
            if ((ad == 1 && max(sdc.setD.A'*(x(:,l+1) - xe)) <= 1) || ...
                    (ad == 0 && (x(:,l+1) - xe)'*sdc.setD.P*(x(:,l+1) - xe) <= 1))
                if (Jnew < Jsub)
                    updates = updates + 1;
                    Uw = Uq;
                    Jsub = Jnew;
                    xf = x(:,end);
                end
            end
        end
    end
    sdc.u(:,k) = Uw(:,1);
    sdc.x(:,k+1) = sdc.modelfun(sdc.x(:,k) - xe,sdc.u(:,k) - ue) + xe;
    if (size(Uw,2) > sdc.Nc)
        Uw = Uw(:,2:end);
    else
        sdc.ncom(k) = sdc.ncom(k) + 1;
        Uw = [Uw(:,2:end) sdc.K*(xf - xe) + ue];
    end
    sdc.J(k) = Jsub;
    sdc.t(k) = toc;
    disp(['iteration: ',num2str(k),', J = ',num2str(sdc.J(k)),' (',num2str(updates),' updates)']);
    sdc.updates = sdc.updates + updates;
    
    x = [sdc.x(:,k+1) zeros(n,size(Uw,2))];
    for i = 1:size(Uw,2)
        x(:,i+1) = sdc.modelfun(x(:,i) - xe,Uw(:,i) - ue) + xe;
    end
    k = k + 1;
end
end