clear, close all;
Vin = 20;
Vd = 0.7;
Z = 80;
Ts = 25e-6;
L = 390e-6;
C = 22e-6;

% equilibrium point
Vo = 10;
mod.xe = [(Vo + Vin + Vd)*Vo/(Vin*Z);Vo];
mod.ue = (Vo + Vd)/(Vo + Vin + Vd);
mod.x0 = zeros(2,1);

% model buck-boost converter
% xp = A*x + B*u + C(x)u + f (f not included)
mod.A0 = [1 -Ts/L;Ts/C 1-Ts/(C*Z)];
mod.B0 = [(Vin + Vd)*Ts/L;0];
mod.B{1} = [0;Ts/L];
mod.B{2} = [-Ts/C;0];

% shift model such that the origin is the equilibrium point
mod.A0 = mod.A0 + [mod.ue'*mod.B{1}';mod.ue'*mod.B{2}'];
mod.B0 = mod.B0 + [mod.xe'*mod.B{1};mod.xe'*mod.B{2}];

% set of allowable states
setX.v = [0 0;2.5 0;2.5 50;0 50]' - mod.xe;
setX.A = vert2con(setX.v')';

% set of allowable inputs
setU.v = [0 1] - mod.ue;
[setU.M, setU.um] = vert2con(setU.v');

%% Domain of attraction method lfl
% pole placement
tic;
lfl.lambda = [0.8,0.9];
lfl.K0 = -place(mod.A0,mod.B0,lfl.lambda);

% determine initial polytope (set P)
[P,~] = eig(mod.A0 + mod.B0*lfl.K0);
lfl.setP.A = (kron(eye(2),[1;-1])*P^(-1))';
lfl.setP.v = con2vert(lfl.setP.A')';

% scale initial polytope (set P)
alpha = sdpvar(1);
cons = [];
for i = 1:size(lfl.setP.v,2)
    cons = [cons;setX.A'*lfl.setP.v(:,i)*alpha <= ones(size(setX.A,2),1)];
    cons = [cons;setU.M*lfl.K0*lfl.setP.v(:,i)*alpha <= setU.um];
end
lfl.t = toc;
options = sdpsettings('solver','linprog','verbose',0);
z = optimize(cons,-alpha,options);
lfl.t = lfl.t + z.solvertime;
lfl.setP.v = lfl.setP.v*value(alpha);
lfl.setP.A = vert2con(lfl.setP.v')';

% determine controller corresponding by initial set P
[lfl.K0,epsilon,t0] = LFLDetContr2(mod,lfl.setP,setU);
lfl.t = lfl.t + t0;
% run algorithm
lfl.s0 = 1;
lfl.sast = 0.05;
lfl.maxiter = 200;
lfl.add_vertices = 0;
lfl.zeta = 2;
[lfl.K,lfl.setP,t0] = doa_quad_model_lfl(lfl.setP,lfl.s0,lfl.sast,...
    lfl.zeta,lfl.maxiter,lfl.add_vertices,setX,setU,mod);
lfl.t = lfl.t + t0;

%% Linear parameter varying MPC
lpv.Nhat = 5;
lpv.p = 1;
lpv.x0 = -mod.xe;
lpv.P = blkdiag(2,1,1);
lpv.Q = blkdiag(2,1,1);
lpv.R = 0.1;
lpv.N = 1000;
[lpv.x,lpv.u,lpv.du,time(3)] = initseq_lpv_mpc_v2(mod,lpv,setX,setU,lfl.setP);
lpv.u = lpv.u(:,1:end-1);
lpv.N = size(lpv.x,2)-1;
t = 0:Ts:Ts*lpv.N;
figure, plot(t,lpv.x(1,:)+mod.xe(1)), hold on, plot(t,lpv.x(2,:)+mod.xe(2)), grid;
figure, stairs(t(1:end-1),lpv.u+mod.ue), grid;
figure, stairs(t(1:end-1),lpv.u,'-o'), grid;
xlabel('time [s]'), ylabel('u');
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(lfl.setP.v'),'alpha',0.35,'color','Orange');
plot(0,0,'r*');
plot(lpv.x(1,:),lpv.x(2,:),'-o','color','Yellow'), grid;
% xlim([-0.2 2.3]), ylim([-11 6]);
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}', '\bf{D}', 'origin', 'x');

%% SD-NMPC online algorithm
sdc.kmax = 200;
sdc.N = lpv.N;
sdc.Ns = 5;
sdc.Nc = 5;
sdc.K = lfl.K;
sdc.setD = lfl.setP;
sdc.Q = blkdiag(1,7);
sdc.R = 0.1;
sdc.x0 = lpv.x0 + mod.xe;
sdc.u = lpv.u + mod.ue;
sdc.modelfun = @(x,u)BuckBoostModel(x,mod,[],u);
sdc.setX = setX;
sdc.setU = setU;
sdc.par.method = 'uniform';
sdc.par.mu = mod.ue;
sdc.par.sigma = 0.1;
sdc.sampfun = @(s,par)sdnmpc_sampling_BBC(s,par);

x(:,1) = sdc.x0;
sdc.ui = sdc.u;
k = 1;
while(k <= sdc.kmax)
    sdc.Jmax(k) = 0;
    for i = 1:sdc.Nc
        if (k+i-1 > sdc.N)
            sdc.ui(:,k+i-1) = sdc.K*(x(:,k+i-1) - mod.xe) + mod.ue;
        end
        x(:,k+i) = BuckBoostModel(x(:,k+i-1) - mod.xe,mod,[],sdc.ui(:,k+i-1) - mod.ue) + mod.xe;
        sdc.Jmax(k) = sdc.Jmax(k) + (x(:,k+i) - mod.xe)'*sdc.Q*(x(:,k+i) - mod.xe) + ...
            (sdc.ui(:,k+i-1) - mod.ue)'*sdc.R*(sdc.ui(:,k+i-1) - mod.ue);
    end
    k = k + 1;
end
sdc.ui = sdc.ui(:,1:sdc.kmax);
sdc.xi = x(:,1:sdc.kmax+1);

sdc = sdnmpc(sdc,mod.xe,mod.ue);
t = 0:Ts:Ts*sdc.kmax;
figure, stairs(t,sdc.x'), grid on;
hold on, plot([0 Ts*sdc.kmax],mod.xe(1)*ones(2,1),'--')
plot([0 Ts*sdc.kmax],mod.xe(2)*ones(2,1),'--')
legend('inductor current [A]', 'output voltage [V]', 'current reference', 'voltage reference');
xlabel('time [s]')
figure, stairs(t,sdc.x(1,:)), grid on;
hold on, stairs(t,sdc.xi(1,:))
plot([0 Ts*sdc.kmax],mod.xe(1)*ones(2,1),'--')
legend('after SD-NMPC', 'before SD-NMPC', 'current reference');
xlabel('time [s]')
ylabel('i_L [A]');
figure, stairs(t,sdc.x(2,:)), grid on;
hold on, stairs(t,sdc.xi(2,:))
plot([0 Ts*sdc.kmax],mod.xe(2)*ones(2,1),'--')
legend('after SD-NMPC', 'before SD-NMPC', 'voltage reference');
xlabel('time [s]')
ylabel('v_o [V]')
figure, plot(t(1:end-1),sdc.J), hold on, plot(t(1:end-1),sdc.Jmax), grid on;
legend('after sd-nmpc', 'before sd-nmpc');
xlabel('time [s]');
figure, stairs(t(1:end-1),sdc.u'), grid on;
hold on, plot([0 Ts*sdc.kmax],mod.ue*ones(2,1),'--')
legend('duty cycle of switch', 'reference duty cycle');
xlabel('time [s]');
figure, plot(t(1:end-1),sdc.t)
figure, stairs(t(1:end-1),[sdc.ui;sdc.u]')
hold on, plot([0 Ts*sdc.kmax],mod.ue*ones(2,1),'--')
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(lfl.setP.v'),'alpha',0.35,'color','Orange');
plot(0,0,'r*');
plot(sdc.x(1,:)-mod.xe(1),sdc.x(2,:)-mod.xe(2),'-','color','Yellow'), grid;
plot(sdc.xi(1,:)-mod.xe(1),sdc.xi(2,:)-mod.xe(2),'-','color','Green'), grid;
% xlim([-0.2 2.3]), ylim([-11 6]);
xlabel('x^{(1)}'), ylabel('x^{(2)}');
legend('\bf{X}', '\bf{D}', 'origin', 'x');

%% efficiency
sdc.Pin = Vin.*sdc.u.*sdc.x(1,1:end-1);
sdc.Pout = sdc.x(2,1:end-1).^2/Z;
% sdc.eta = sdc.Pout./sdc.Pin;
sdc.eta = cumsum(sdc.Pout)./cumsum(sdc.Pin);
figure, plot(t(1:end-1),sdc.eta), grid;
hold on, plot(t(1:end-1),ones(max(size(t))-1,1),'--');

%% PID controller
if (Vo == 10)
    load('FRD_BBC_Vo_10.mat');
elseif (Vo == 40)
    load('FRD_BBC_Vo_40.mat');
end
pidc.estsys = estsys;
figure, bode(pidc.estsys), grid;
clear estsys;
if (Vo == 10)
    pidc.Kp = 104;
    pidc.phimax = 70;
    pidc.alpha = (1 + sind(pidc.phimax))/(1 - sind(pidc.phimax));
    pidc.omegamax = 2e4;
    pidc.omegaL = pidc.omegamax/sqrt(pidc.alpha);
elseif(Vo == 40)
    pidc.Kp = 7;
    pidc.phimax = 75;
    pidc.alpha = (1 + sind(pidc.phimax))/(1 - sind(pidc.phimax));
    pidc.omegamax = 1e4;
    pidc.omegaL = pidc.omegamax/sqrt(pidc.alpha);
end
s = tf('s');
pidc.C = pidc.Kp;
pidc.C = pidc.C/s;
figure, margin(pidc.C*pidc.estsys), grid;
pidc.C = pidc.C*(s/pidc.omegaL + 1)^2/(s/(pidc.alpha*pidc.omegaL) + 1)^2;
hold on, bode(pidc.C*pidc.estsys);
% pidc.C = pidc.C*((s/pidc.w1)^2 + s/(pidc.Q1*pidc.w1) + 1)/((s/pidc.w2)^2 + s/(pidc.Q2*pidc.w2) + 1);
pidc.G = pidc.C*pidc.estsys;
figure, margin(pidc.G), grid;
figure, bode(pidc.C/(1 + pidc.G)), grid;
figure, bode(pidc.G/(1 + pidc.G)), grid;
figure, step(tfest(pidc.G,20)/(1 + tfest(pidc.G,20))), grid;

%% Comparison between PID control and SD-NMPC control
if (Vo == 10)
    load('PID_before_SDC_BBC_Vo_10.mat');
    initdata = data;
    load('PID_SDC_BBC_Vo_10.mat');
    load('obj_SDC_BBC_Vo_10.mat');
elseif (Vo == 40)
    load('PID_before_SDC_BBC_Vo_40.mat');
    initdata = data;
    load('PID_SDC_BBC_Vo_40.mat');
    load('obj_SDC_BBC_Vo_40.mat');
end
figure;
if (Vo == 10)
    sl = 61;
elseif (Vo == 40)
    sl = 201;
end
stairs(data{3}.Values.Time(1:sl),data{3}.Values.Data(1:sl)), grid on;
hold on, stairs(initdata{3}.Values.Time(1:sl),initdata{3}.Values.Data(1:sl));
plot(data{3}.Values.Time(1:sl),mod.xe(1)*ones(size(data{3}.Values.Time(1:sl),1),1),'--');
legend('after SD-NMPC','before SD-NMPC','current reference');
xlabel('time [s]'), ylabel('i_L [A]');
figure;
if (Vo == 10)
    sl = 101;
elseif (Vo == 40)
    sl = 401;
end
stairs(data{4}.Values.Time(1:sl),data{4}.Values.Data(1:sl)), grid on;
hold on, stairs(initdata{4}.Values.Time(1:sl),initdata{4}.Values.Data(1:sl));
plot(data{4}.Values.Time(1:sl),mod.xe(2)*ones(size(data{4}.Values.Time(1:sl),1),1),'--');
legend('after SD-NMPC','before SD-NMPC','voltage reference');
xlabel('time [s]'), ylabel('v_o [V]');
if (Vo == 10)
    ylim([0 11]);
elseif (Vo == 40)
    ylim([0 43]);
end
figure;
if (Vo == 10)
    sl = 41;
elseif (Vo == 40)
    sl = 161;
end
stairs(data{8}.Values.Time(1:sl),data{8}.Values.Data(1:sl)), grid on;
hold on, plot(data{8}.Values.Time(1:sl),mod.ue*ones(size(data{8}.Values.Time(1:sl),1),1),'--');
legend('duty cycle of the switch','reference duty cycle');
xlabel('time [s]'), ylabel('d');
figure;
if (Vo == 10)
    sl = 21;
elseif (Vo == 40)
    sl = 161;
end
plot(time(1:sl),J(1:sl)), hold on, plot(time(1:sl),Jinit(1:sl)), grid on;
legend('after SD-NMPC', 'before SD-NMPC');
xlabel('time [s]'), ylabel('J');
if (Vo == 10)
    xlim([0 5e-4]);
end
figure;
subplot(1,2,1), stairs(data{1}.Values.Time,data{1}.Values.Data), grid;
hold on, stairs(data{3}.Values.Time,data{3}.Values.Data);
legend('PID controller', 'SD-NMPC controller');
xlabel('time [s]'), ylabel('i_L [A]');
if (Vo == 40)
    xlim([0 0.01]);
end
subplot(1,2,2), stairs(data{2}.Values.Time,data{2}.Values.Data), grid;
hold on, stairs(data{4}.Values.Time,data{4}.Values.Data);
legend('PID controller', 'SD-NMPC controller');
xlabel('time [s]'), ylabel('v_o [V]');
if (Vo == 10)
    ylim([0 11]);
else
    ylim([0 44]);
end

Imin = zeros(size(data{3}.Values.Time,1),1);
Imax = zeros(size(data{3}.Values.Time,1)-1,1);
Vmin = zeros(size(data{3}.Values.Time,1)-1,1);
Vmax = zeros(size(data{3}.Values.Time,1)-1,1);
for i = 1:size(data{3}.Values.Time,1)-1
    Imax(i) = Imin(i) + Vin*data{8}.Values.Data(i)*Ts/L;
    Imin(i+1) = Imax(i) - (data{4}.Values.Data(i+1) + Vd)*(1 - data{8}.Values.Data(i))*Ts/L;
    Vmax(i) = L*(Imax(i)-Imin(i))/(data{8}.Values.Data(i)*Ts);
    Vmin(i) = L*(Imin(i+1)-Imax(i))/((1 - data{8}.Values.Data(i))*Ts);
    if (isnan(Vmin(i)))
        Vmin(i) = 0;
    end
end
figure;
stairs(data{3}.Values.Time(1:end-1),[Imin(2:end) Imax],'--'), grid on;
hold on, stairs(data{3}.Values.Time(1:end-1),data{3}.Values.Data(2:end));
if (Vo == 10)
    sl = 41;
elseif (Vo == 40)
    sl = 120;
end
figure;
stairs(data{3}.Values.Time(1:sl),[Vmin(1:sl) Vmax(1:sl)],'--'), grid on;
hold on, stairs(data{3}.Values.Time(1:sl),L*diff(data{3}.Values.Data(1:sl+1))/Ts)
legend('V_{min}','V_{max}','v_L');
xlabel('time [s]'), ylabel('inductor voltage [V]');
t = zeros(size(Vmin,1) + size(Vmax,1),1);
VL = zeros(size(Vmin,1) + size(Vmax,1),1);
for i = 1:size(Vmin,1)
    t(2*i-1) = (i - 1)*Ts;
    t(2*i) = (i - 1 + data{8}.Values.Data(i))*Ts;
    VL(2*i-1:2*i) = [Vmax(i);Vmin(i)];
end
figure, stairs(t(1:2*sl),VL(1:2*sl),'color','Green'), grid on;
hold on, stairs(data{3}.Values.Time(1:sl),L*diff(data{3}.Values.Data(1:sl+1))/Ts,'color','Blue')
legend('inductor voltage','average inductor voltage');
xlabel('time [s]'), ylabel('v_L [V]');
if (Vo == 10)
    xlim([0 1e-3]);
end