clear, close all;
Vin = 220;
p = 6;
Rs = 1.5;
lambda_pm = 0.05;
Lq = 1.1e-4;
Ld = 3.5e-4;
JL = 0.019;
DL = 0.12;
TL0 = 30;
Ts = 1e-4;
omega0 = 140;

% equilibrium point of the pmsm
mod.xe(1,1) = 0;
mod.xe(2,1) = 2*(DL*omega0 + TL0)/(sqrt(3)*p*lambda_pm);
mod.xe(3,1) = omega0;
mod.ue(1,1) = -p*Lq*omega0*mod.xe(2)/3;
mod.ue(2,1) = (Rs*mod.xe(2) + sqrt(3)*p*lambda_pm*omega0)/3;

% model pmsm
mod.A0 = [1-Rs*Ts/Ld 0 0;0 1-Rs*Ts/Lq -sqrt(3)*p*lambda_pm*Ts/Lq;...
    0 sqrt(3)*p*lambda_pm*Ts/(2*JL) 1-DL*Ts/JL];
mod.A{1} = zeros(3);
mod.A{1}(3,2) = p*Lq*Ts/Ld;
mod.A{2} = zeros(3);
mod.A{2}(3,1) = -p*Ld*Ts/Lq;
mod.A{3} = zeros(3);
mod.A{3}(1,2) = p*(Ld-Lq)*Ts/(2*JL);
mod.B0 = [3*Ts/Ld 0;0 3*Ts/Lq;0 0];

% shift model such that the origin is the equilibrium point
mod.A0 = mod.A0 + [mod.xe'*mod.A{1};mod.xe'*mod.A{2};mod.xe'*mod.A{3}] + ...
    [mod.xe'*mod.A{1}';mod.xe'*mod.A{2}';mod.xe'*mod.A{3}'];

% set of allowable states
% setX.v = [-210 -210 0;210 -210 0;-210 210 0;210 210 0;...
%     -210 -210 150;210 -210 150;-210 210 150;210 210 150]' - mod.xe;
setX.v = interval2vert(kron([-1 1],[210;210;150]))' - mod.xe;
setX.A = vert2con(setX.v')';

% set of allowable inputs
th = linspace(0,2*pi,9);
setU.v = (Vin/sqrt(3))*[cos(th(1:end-1));sin(th(1:end-1))] - mod.ue;
[setU.M, setU.um] = vert2con(setU.v');

%% Linear feedback using quadratic Lyapunov function
% determine initial proposed DOA
P = sdpvar(3);
Q = sdpvar(3);
R = sdpvar(2);
cons = [];
for i = 1:size(setX.A,2)
    cons = [cons;[1 setX.A(:,i)';setX.A(:,i) P] >= eps];
end
cons = [cons;[mod.A0'*P*mod.A0-P+Q mod.A0'*P*mod.B0;...
    mod.B0'*P*mod.A0 mod.B0'*P*mod.B0+R] >= 0];
cons = [cons;P >= eps;Q >= 0; R >= eps];
obj = trace(P);
options = sdpsettings('solver','mosek','verbose',0);
z = optimize(cons,obj,options);
lfq.t = z.solvertime;
lfq.setD.P = value(P);

% determine polytope P
N = 12;
th = linspace(0,pi,N+1);
xi = [];
for i = 1:N
    for j = 1:N
        xi = [xi lfq.setD.P^(-0.5)*[cos(th(i));sin(th(i))*cos(2*th(j));sin(th(i))*sin(2*th(j))]];
    end
end
lfq.setP.v = xi;
lfq.setP.A = vert2con(xi')';

gamma = 0;
for i = 1:size(lfq.setP.A,2)
    gamma = max(gamma,lfq.setP.A(:,i)'*lfq.setD.P^(-1)*lfq.setP.A(:,i));
end
lfq.setD.P = lfq.setD.P*value(gamma);
figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
hold on, plot(Polyhedron(lfq.setP.v'),'alpha',0.35,'color','Orange');
Ellipse_plot(lfq.setD.P,zeros(3,1));
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
legend('\bf{X}', '\bf{P}', '\partial\bf{D}_0');
xlim([-220 220]), ylim([-35 35]), zlim([-11 11]);

% apply algorithm in order to find feedback gain and DOA
lfq.delta = 0.01;
lfq.imax = 20;
[lfq.K,lfq.setP,t0] = doa_quad_model_lfq(mod,lfq.setD.P,...
    lfq.setP,setU,lfq.delta,lfq.imax);
lfq.t = lfq.t + t0;
gamma = 0;
for i = 1:size(lfq.setP.A,2)
    gamma = max(gamma,lfq.setP.A(:,i)'*lfq.setD.P^(-1)*lfq.setP.A(:,i));
end
lfq.setD.P = lfq.setD.P*gamma;

%% Successive linearization
sln.x0 = -mod.xe;
sln.n = size(mod.A0,1);
sln.m = size(mod.B0,2);
sln.N = 2e4;
sln.Nc = 5;
sln.x(:,1) = sln.x0;
sln.Q = blkdiag(1,1,10);
sln.setX = setX;
sln.setU = setU;
sln.modelfun = @(x,u)PMSMModel(x,mod,[],u);
[sln.x,sln.u,sln.N,time(2)] = initseq_sln(sln,mod,lfq.setD.P,1);
t = 0:Ts:Ts*sln.N;
figure, stairs(t,sln.x(1,:)+mod.xe(1)), hold on, stairs(t,sln.x(2,:)+mod.xe(2)), grid;
stairs(t,sln.x(3,:)+mod.xe(3))
figure, stairs(t(1:end-1),sln.u(1,:)+mod.ue(1)), hold on, stairs(t(1:end-1),sln.u(2,:)+mod.ue(2)), grid;
figure, stairs(t(1:end-1),sln.u(1,:)), hold on, stairs(t(1:end-1),sln.u(2,:)), grid;
legend('u^{(1)}', 'u^{(2)}');
xlabel('time [s]')
% figure, plot(Polyhedron(setX.v'),'alpha',0.35,'color','Blue');
figure, Ellipse_plot(lfq.setD.P,zeros(3,1),[],'Yellow','fill');
hold on, plot3(sln.x(1,:),sln.x(2,:),sln.x(3,:),'-','color','Green'), grid on;
xlabel('x^{(1)}'), ylabel('x^{(2)}'), zlabel('x^{(3)}');
legend('\bf{D}', 'x');

%% SD-NMPC online algorithm
sdc.kmax = 1e4;
sdc.N = sln.N;
sdc.Ns = 5;
sdc.Nc = 5;
sdc.K = lfq.K;
sdc.setD = lfq.setD;
sdc.Q = blkdiag(1,1,10);
sdc.R = blkdiag(1,1);
sdc.rho = 0;
sdc.x0 = sln.x0 + mod.xe;
sdc.u = sln.u + mod.ue;
sdc.modelfun = @(x,u)PMSMModel(x,mod,[],u);
sdc.setX = setX;
sdc.setU = setU;
sdc.par.method = 'truncated normal';
sdc.par.mu = mod.ue;
sdc.par.sigma = [0.05;0.01];
sdc.sampfun = @(s,par)sdnmpc_sampling_PMSM(s,par,Vin);

sdc.xi = [sdc.x0 zeros(size(mod.xe,1),sdc.kmax)];
sdc.ui = sdc.u;
k = 1;
while(k <= sdc.kmax)
    sdc.Jmax(k) = 0;
    for i = 1:sdc.Nc
        if (k+i-1 > sdc.N)
            sdc.ui(:,k+i-1) = sdc.K*(sdc.xi(:,k+i-1) - mod.xe) + mod.ue;
        end
        sdc.xi(:,k+i) = PMSMModel(sdc.xi(:,k+i-1) - mod.xe,mod,[],sdc.ui(:,k+i-1) - mod.ue) + mod.xe;
        sdc.Jmax(k) = sdc.Jmax(k) + (sdc.xi(:,k+i) - mod.xe)'*sdc.Q*(sdc.xi(:,k+i) - mod.xe) + ...
            (sdc.ui(:,k+i-1) - mod.ue)'*sdc.R*(sdc.ui(:,k+i-1) - mod.ue);
    end
    k = k + 1;
end
sdc.ui = sdc.ui(:,1:sdc.kmax);
sdc.xi = sdc.xi(:,1:sdc.kmax+1);

x(:,1) = sdc.x0;
if (sdc.N < sdc.Nc)
    for i = 1:sdc.Nc
        if (i > sdc.N)
            sdc.u(:,i) = sdc.K*(x(:,i) - mod.xe) + mod.ue;
        end
        x(:,i+1) = PMSMModel(x(:,i) - mod.xe,mod,[],sdc.u(:,i) - mod.ue);
        x(:,i+1) = x(:,i+1) + mod.xe;
    end
end
sdc = sdnmpc(sdc,mod.xe,mod.ue);
t = 0:Ts:Ts*sdc.kmax;
figure;
subplot(2,1,1), stairs(t,[sdc.x(1,:);sdc.xi(1,:)]'), grid;
hold on, plot([0 Ts*sdc.kmax],mod.xe(1)*ones(2,1),'--');
legend('after SD-NMPC', 'before SD-NMPC', 'reference current');
xlabel('time [s]'), ylabel('i_d [A]');
subplot(2,1,2), stairs(t,[sdc.x(2,:);sdc.xi(2,:)]'), grid;
hold on, plot([0 Ts*sdc.kmax],mod.xe(2)*ones(2,1),'--');
legend('after SD-NMPC', 'before SD-NMPC', 'reference current');
xlabel('time [s]'), ylabel('i_q [A]');
figure, plot(t(1:end-1),sdc.J), hold on, plot(t(1:end-1),sdc.Jmax), grid;
legend('after SD-NMPC','before SD-NMPC');
xlabel('time [s]');
figure, stairs(t,[sdc.x(3,:);sdc.xi(3,:)]'), grid;
hold on, plot([0 Ts*sdc.kmax],mod.xe(3)*ones(2,1),'--');
legend('after SD-NMPC', 'before SD-NMPC', 'reference speed');
xlabel('time [s]'), ylabel('\omega [rad/s]');
ylim([0 160]);
figure;
subplot(2,1,1), stairs(t(1:end-1),[sdc.u(1,:);sdc.ui(1,:)]'), grid;
hold on, plot([0 Ts*sdc.kmax],mod.ue(1)*ones(2,1),'--');
legend('after SD-NMPC', 'before SD-NMPC', 'reference voltage');
xlabel('time [s]'), ylabel('v_d [A]');
subplot(2,1,2), stairs(t(1:end-1),[sdc.u(2,:);sdc.ui(2,:)]'), grid;
hold on, plot([0 Ts*sdc.kmax],mod.ue(2)*ones(2,1),'--');
legend('after SD-NMPC', 'before SD-NMPC', 'reference voltage');
xlabel('time [s]'), ylabel('v_q [A]');

%% PID controller
close all;
% load('FRD_PMSM_omega_r.mat');
s = tf('s');
pidc.sysd = (3/Ld)/(s + Rs/Ld);
figure, margin(pidc.sysd), grid;
pidc.Kp1 = 1.2;
pidc.omega_i1 = 4e3;
pidc.Cd = pidc.Kp1*(1 + pidc.omega_i1/s);
pidc.Gd = pidc.Cd*pidc.sysd;
pidc.sysq = (3/Lq)/(s + Rs/Lq);
figure, margin(pidc.sysq), grid;
pidc.Kp2 = 0.58;
pidc.omega_i2 = 4e3;
pidc.Cq = pidc.Kp2*(1 + pidc.omega_i2/s);
pidc.Gq = pidc.Cq*pidc.sysq;
pidc.sysw = sqrt(3)*p*lambda_pm/(2*JL)*1/(s + DL/JL);
figure, margin(pidc.sysw), grid;
% pidc.sysw = estsys;
% clear estsys;
pidc.Kp3 = 1;
pidc.omega_i3 = 8;
pidc.Cw = pidc.Kp3*(1 + pidc.omega_i3/s);
pidc.Gw = pidc.Cw*pidc.sysw;
figure, margin(pidc.Gd), grid;
figure, margin(pidc.Gq), grid;
figure, margin(pidc.Gw), grid;
figure, bode(pidc.Cw/(1 + pidc.Gw)), grid;
figure, stepplot(pidc.Gw/(1 + pidc.Gw)), grid;

%%
load('PMSM_simulation_140_after_SD_NMPC_splitted.mat');
figure(1)
subplot(2,1,1), stairs(data{4}.Values.Time,data{4}.Values.Data(:,1)), grid;
subplot(2,1,2), stairs(data{4}.Values.Time,data{4}.Values.Data(:,2)), grid;
figure(2)
stairs(data{14}.Values.Time,data{14}.Values.Data(:,1)), grid;
figure(3)
plot(0:Ts:(max(size(J))-1)*Ts,J), grid;
figure(4)
subplot(2,1,1), stairs(data{6}.Values.Time,data{6}.Values.Data(:,1)), grid;
subplot(2,1,2), stairs(data{6}.Values.Time,data{6}.Values.Data(:,2)), grid;

load('PMSM_simulation_140_before_SD_NMPC_splitted.mat');
figure(1)
subplot(2,1,1), hold on, stairs(data{4}.Values.Time,data{4}.Values.Data(:,1));
plot([data{4}.Values.Time(1), data{4}.Values.Time(end)],mdl.xf(1)*ones(2,1),'-');
xlabel('time [s]'), ylabel('i_d [A]');
legend('after SD-NMPC', 'before SD-NMPC', 'reference current');
subplot(2,1,2), hold on, stairs(data{4}.Values.Time,data{4}.Values.Data(:,2));
plot([data{4}.Values.Time(1), data{4}.Values.Time(end)],mdl.xf(2)*ones(2,1),'-');
xlabel('time [s]'), ylabel('i_q [A]');
legend('after SD-NMPC', 'before SD-NMPC', 'reference current');
figure(2)
hold on, stairs(data{14}.Values.Time,data{14}.Values.Data(:,1))
plot([data{14}.Values.Time(1), data{14}.Values.Time(end)],mdl.xf(3)*ones(2,1),'--');
ylim([-10 150]);
xlabel('time [s]'), ylabel('\omega_r [rad/s]');
legend('after SD-NMPC', 'before SD-NMPC', 'reference speed');
figure(3)
hold on, plot(0:Ts:(max(size(J))-1)*Ts,J);
xlabel('time [s]');
legend('after SD-NMPC', 'before SD-NMPC')
figure(4)
subplot(2,1,1), hold on, stairs(data{6}.Values.Time,data{6}.Values.Data(:,1));
plot([data{6}.Values.Time(1), data{6}.Values.Time(end)],mdl.uf(1)*ones(2,1),'--');
xlabel('time [s]'), ylabel('v_d [V]');
legend('after SD-NMPC', 'before SD-NMPC', 'reference voltage');
subplot(2,1,2), hold on, stairs(data{6}.Values.Time,data{6}.Values.Data(:,2));
plot([data{6}.Values.Time(1), data{6}.Values.Time(end)],mdl.uf(2)*ones(2,1),'--');
xlabel('time [s]'), ylabel('v_q [V]');
legend('after SD-NMPC', 'before SD-NMPC', 'reference voltage');



%% PID controller results
% load('PID_PMSM_data_140.mat');
load('PID_PMSM_data_80.mat');
figure;
subplot(2,1,1), stairs(data{4}.Values.Time,data{4}.Values.Data(:,1)), grid;
hold on, stairs(data{5}.Values.Time,data{5}.Values.Data(:,1));
% xlim([0 0.8]);
xlim([0 0.5]);
xlabel('time [s]'), ylabel('i_d [A]'), legend('SD-NMPC controller', 'PI controllers');
subplot(2,1,2), stairs(data{4}.Values.Time,data{4}.Values.Data(:,2)), grid;
hold on, stairs(data{5}.Values.Time,data{5}.Values.Data(:,2));
% xlim([0 0.8]);
xlim([0 0.5]);
xlabel('time [s]'), ylabel('i_q [A]'), legend('SD-NMPC controller', 'PI controllers');
figure, stairs(data{14}.Values.Time,data{14}.Values.Data), grid;
hold on, stairs(data{13}.Values.Time,data{13}.Values.Data);
% xlim([0 0.8]), ylim([0 149]);
xlim([0 0.8]), ylim([-10 89]);
xlabel('time [s]'), ylabel('\omega_r [rad/s]'), legend('SD-NMPC controller', 'PI controllers');

%%
close all;
load('PID_SDC_PMSM_140_fast_after.mat');
load('obj_SDC_VSI_140_fast.mat');
figure(1);
subplot(2,1,1)
stairs(data{4}.Values.Time,data{4}.Values.Data(:,1)), grid on;
xlim([0 1.2]);
subplot(2,1,2)
stairs(data{4}.Values.Time,data{4}.Values.Data(:,2)), grid on;
xlim([0 0.7]);

figure(2);
stairs(data{14}.Values.Time,data{14}.Values.Data), grid on;
xlim([0 1.4]);

figure(3);
subplot(2,1,1)
stairs(data{6}.Values.Time,data{6}.Values.Data(:,1)), grid on;
xlim([0 1.2]);
subplot(2,1,2)
stairs(data{6}.Values.Time,data{6}.Values.Data(:,2)), grid on;
xlim([0 1]);

figure(4);
stairs(time,[J Jinit]), grid on;
xlim([0 0.8]);
legend('after SD-NMPC','before SD-NMPC');
xlabel('time [s]'), ylabel('J');

figure;
subplot(3,1,1)
stairs(data{1}.Values.Time(1:2695),data{1}.Values.Data(1:2695,1)), grid on;
subplot(3,1,2)
stairs(data{1}.Values.Time(1:2695),data{1}.Values.Data(1:2695,1)), grid on;
subplot(3,1,3)
stairs(data{1}.Values.Time(1:2695),data{1}.Values.Data(1:2695,1)), grid on;

figure;
subplot(3,1,1)
stairs(data{11}.Values.Time,data{11}.Values.Data(:,1)), grid on;
xlim([0 0.25]);
xlabel('time [s]'), ylabel('i_a [A]');
subplot(3,1,2)
stairs(data{11}.Values.Time,data{11}.Values.Data(:,1)), grid on;
xlim([0 0.25]);
xlabel('time [s]'), ylabel('i_b [A]');
subplot(3,1,3)
stairs(data{11}.Values.Time,data{11}.Values.Data(:,1)), grid on;
xlim([0 0.25]);
xlabel('time [s]'), ylabel('i_c [A]');

load('PID_SDC_PMSM_140_fast_before.mat');
figure(1);
subplot(2,1,1)
hold on, stairs(data{4}.Values.Time,data{4}.Values.Data(:,1));
plot(data{4}.Values.Time,mod.xe(1)*ones(size(data{4}.Values.Time,1),1),'--');
legend('after SD-NMPC','before SD-NMPC','reference current');
xlabel('time [s]'), ylabel('i_d [A]');
subplot(2,1,2)
hold on, stairs(data{4}.Values.Time,data{4}.Values.Data(:,2));
plot(data{4}.Values.Time,mod.xe(2)*ones(size(data{4}.Values.Time,1),1),'--');
legend('after SD-NMPC','before SD-NMPC','reference current');
xlabel('time [s]'), ylabel('i_q [A]');

figure(2);
hold on, stairs(data{14}.Values.Time,data{14}.Values.Data);
plot(data{14}.Values.Time,mod.xe(3)*ones(size(data{14}.Values.Time,1),1),'--');
legend('after SD-NMPC','before SD-NMPC','reference speed');
xlabel('time [s]'), ylabel('\omega_r [A]')
ylim([0 150]);

figure(3);
subplot(2,1,1)
hold on, stairs(data{6}.Values.Time,data{6}.Values.Data(:,1));
plot(data{6}.Values.Time,mod.ue(1)*ones(size(data{6}.Values.Time,1),1),'--');
legend('after SD-NMPC','before SD-NMPC','reference voltage');
xlabel('time [s]'), ylabel('v_d [V]');
subplot(2,1,2)
hold on, stairs(data{6}.Values.Time,data{6}.Values.Data(:,2));
plot(data{6}.Values.Time,mod.ue(2)*ones(size(data{6}.Values.Time,1),1),'--');
legend('after SD-NMPC','before SD-NMPC','reference voltage');
xlabel('time [s]'), ylabel('v_q [V]');